"""Postprocessing option computes overall consumption, production,self-consumption and injection as well as selfconsumption rate and autarky rate."""

import os
from dataclasses import dataclass
from typing import Any, List

import pandas as pd
from dataclasses_json import dataclass_json

from building_sizer.kpi_config import KPIConfig
from hisim.component import ComponentOutput
from hisim.loadtypes import ComponentType, InandOutputType, LoadTypes
from hisim.simulationparameters import SimulationParameters


@dataclass_json
@dataclass()
class FuelCost:

    """ Defines the fuel costs in terms of euros and co2. """

    electricity_consumption_in_euro_per_kwh: float = 0.35
    electricity_injection_in_euro_per_kwh: float = 0.2
    electricity_consumption_in_kg_co2_per_kwh: float = 0.2
    oil_consumption_in_euro_per_kwh: float = 0.17
    oil_consumption_in_kg_co2_per_kwh: float = 0.3


def compute_kpis(results: pd.DataFrame, all_outputs: List[ComponentOutput], simulation_parameters: SimulationParameters) -> Any:  # noqa: MC0001
    """Calculation of several KPIs."""
    results['consumption'] = 0
    results['production'] = 0
    results['battery_charge'] = 0
    results['battery_discharge'] = 0
    results['storage'] = 0
    index: int
    output: ComponentOutput

    electricity_price_consumption = pd.Series()
    electricity_price_injection = pd.Series()
    self_consumption = pd.Series()

    price_config = FuelCost()

    # replace that loop by searching for flags -> include also battery things and hydrogen things
    # flags for Postprocessing: cp.ComponentOutput.postprocessing_flag -> loadtpyes.InandOutputType :
    # Consumption, Production, StorageContent, ChargeDischarge
    # CHARGE_DISCHARGE from battery has + and - sign and is production and consumption both in one output
    # heat production of heat pump has - sign in summer, either separate or take absolute value
    # flags for ComponentTypes: cp.ComponentOutput.component_type
    # flags for LoadTypes: cp.ComponentOutput.load_type
    # flags for Units: cp.ComponentOutput.unit

    for index, output in enumerate(all_outputs):

        if output.postprocessing_flag is not None:
            if InandOutputType.ELECTRICITY_PRODUCTION in output.postprocessing_flag:
                results['production'] = results['production'] + results.iloc[:, index]

            elif (
                    InandOutputType.ELECTRICITY_CONSUMPTION_EMS_CONTROLLED in output.postprocessing_flag) \
                    or InandOutputType.ELECTRICITY_CONSUMPTION_UNCONTROLLED in output.postprocessing_flag:
                results['consumption'] = results['consumption'] + results.iloc[:, index]

            elif InandOutputType.STORAGE_CONTENT in output.postprocessing_flag:
                results['storage'] = results['storage'] + results.iloc[:, index]

            elif InandOutputType.CHARGE_DISCHARGE in output.postprocessing_flag:
                if ComponentType.BATTERY in output.postprocessing_flag:
                    results["battery_charge"] = results["battery_charge"] + results.iloc[:, index].clip(lower=0)
                    results["battery_discharge"] = results["battery_discharge"] - results.iloc[:, index].clip(upper=0)
                elif ComponentType.CAR_BATTERY in output.postprocessing_flag:
                    results["consumption"] = results["consumption"] + results.iloc[:, index].clip(lower=0)

            elif LoadTypes.PRICE in output.postprocessing_flag:
                if InandOutputType.ELECTRICITY_CONSUMPTION in output.postprocessing_flag:
                    electricity_price_consumption = results.iloc[:, index]
                elif InandOutputType.ELECTRICITY_INJECTION in output.postprocessing_flag:
                    electricity_price_injection = results.iloc[:, index]
                else:
                    continue

        else:
            continue

    # sum over time make it more clear and better
    consumption_sum = results['consumption'].sum() * simulation_parameters.seconds_per_timestep / 3.6e6
    production_sum = results['production'].sum() * simulation_parameters.seconds_per_timestep / 3.6e6

    if production_sum > 0:
        # account for battery
        production_with_battery = results['production'] + results['battery_discharge']
        consumption_with_battery = results['consumption'] + results['battery_charge']

        # evaluate injection and sum over time
        injection = (production_with_battery - consumption_with_battery)
        injection_sum = injection[injection > 0].sum() * simulation_parameters.seconds_per_timestep / 3.6e6

        # evaluate self consumption and immidiately sum over time
        self_consumption = pd.concat((results["production"][results["production"] <= consumption_with_battery],
        consumption_with_battery[consumption_with_battery < results["production"]])).groupby(level=0).sum()
        self_consumption_sum = self_consumption.sum() * simulation_parameters.seconds_per_timestep / 3.6e6

        battery_losses = 0
    else:
        self_consumption_sum = 0
        injection_sum = 0
        battery_losses = 0
    h2_system_losses = 0  # explicitly compute that

    # Electricity Price
    price = 0
    co2 = 0
    if production_sum > 0:
        # evaluate electricity price
        if not electricity_price_injection.empty:
            price = price - (injection[injection > 0] * electricity_price_injection[injection > 0]).sum() \
                * simulation_parameters.seconds_per_timestep / 3.6e6
        else:
            price = price - injection_sum * price_config.electricity_injection_in_euro_per_kwh
        self_consumption_rate = 100 * (self_consumption_sum / production_sum)
        autarky_rate = 100 * (self_consumption_sum / consumption_sum)
    else:
        self_consumption_rate = 0
        autarky_rate = 0

    if not electricity_price_consumption.empty:
        # substract self consumption from consumption for bill calculation
        if not self_consumption.empty:
            results['consumption'] = results['consumption'] - self_consumption
        price = price + (results['consumption'] * electricity_price_consumption).sum() \
            * simulation_parameters.seconds_per_timestep / 3.6e6
    else:
        price = price + (consumption_sum - self_consumption_sum) * price_config.electricity_consumption_in_euro_per_kwh
    co2 = co2 + (consumption_sum - self_consumption_sum) * price_config.electricity_consumption_in_kg_co2_per_kwh

    # initilize lines for report
    lines: List = []
    lines.append(f"Consumption: {consumption_sum:4.0f} kWh")
    lines.append(f"Production: {production_sum:4.0f} kWh")
    lines.append(f"Self-Consumption: {self_consumption_sum:4.0f} kWh")
    lines.append(f"Injection: {injection_sum:4.0f} kWh")
    lines.append(f"Battery losses: {battery_losses:4.0f} kWh")
    lines.append(f"Battery content: {0:4.0f} kWh")
    lines.append(f"Hydrogen system losses: {h2_system_losses:4.0f} kWh")
    lines.append(f"Hydrogen storage content: {0:4.0f} kWh")
    lines.append(f"Autarky Rate: {autarky_rate:3.1f} %")
    lines.append(f"Self Consumption Rate: {self_consumption_rate:3.1f} %")
    lines.append(f"Price paid for electricity: {price:3.0f} EUR")
    lines.append(f"CO2 emitted due to electricity use: {co2:3.0f} kg")

    # initialize json interface to pass kpi's to building_sizer
    kpi_config = KPIConfig(self_consumption_rate=self_consumption_rate, autarky_rate=autarky_rate,
    injection=injection_sum, economic_cost=price, co2_cost=co2)

    pathname = os.path.join(simulation_parameters.result_directory, "kpi_config.json")
    config_file_written = kpi_config.to_json()  # type: ignore
    with open(pathname, 'w', encoding="utf-8") as outfile:
        outfile.write(config_file_written)

    return lines



# Generic/Built-in
from dataclasses import dataclass
from typing import Any
from typing import List

from dataclasses_json import dataclass_json

from hisim import component as cp
from hisim import dynamic_component
from hisim import loadtypes as lt
from hisim import utils
from hisim.simulationparameters import SimulationParameters

__authors__ = "Maximilian Hillen"
__copyright__ = "Copyright 2021, the House Infrastructure Project"
__credits__ = ["Noah Pflugradt"]
__license__ = "MIT"
__version__ = "0.1"
__maintainer__ = "Maximilian Hillen"
__email__ = "maximilian.hillen@rwth-aachen.de"
__status__ = "development"


class L2GenericEnergyManagementSystem(dynamic_component.DynamicComponent):
    """
    Surplus electricity controller - time step based. 
    Iteratively goes through hierachy of devices given by
    source weights of components and passes available surplus
    electricity to each device. Needs to be configured with 
    dynamic In- and Outputs.
    """
    # Inputs

    ElectricityToElectrolyzerUnused = "ElectricityToElectrolyzerUnused"

    # Outputs
    ElectricityToElectrolyzerTarget = "ElectricityToElectrolyzerTarget"

    ElectricityToOrFromGrid = "ElectricityToOrFromGrid"
    TotalElectricityConsumption = "TotalElectricityConsumption"
    FlexibleElectricity = "FlexibleElectricity"
    BuildingTemperatureModifier = "BuildingTemperatureModifier"
    StorageTemperatureModifier = "StorageTemperatureModifier"

    CheckPeakShaving = "CheckPeakShaving"

    @utils.measure_execution_time
    def __init__(self, my_simulation_parameters: SimulationParameters, strategy: str = "optimize_own_consumption",
                 # strategy=["optimize_own_consumption","peak_shaving_from_grid", "peak_shaving_into_grid","seasonal_storage"]
                 limit_to_shave: float = 0):
        self.my_component_inputs: List[dynamic_component.DynamicConnectionInput] = []
        self.my_component_outputs: List[dynamic_component.DynamicConnectionOutput] = []
        super().__init__(my_component_inputs=self.my_component_inputs, my_component_outputs=self.my_component_inputs,
                         name="L2EMSElectricityController", my_simulation_parameters=my_simulation_parameters)

        self.strategy = strategy
        self.limit_to_shave = limit_to_shave
        self.electricity_to_electrolyzer_unused: cp.ComponentInput = self.add_input(object_name=self.component_name,
                                                                                    field_name=self.ElectricityToElectrolyzerUnused,
                                                                                    load_type=lt.LoadTypes.ELECTRICITY, unit=lt.Units.WATT,
                                                                                    mandatory=False)

        # Outputs
        self.electricity_to_or_from_grid: cp.ComponentOutput = self.add_output(object_name=self.component_name,
                                                                               field_name=self.ElectricityToOrFromGrid,
                                                                               load_type=lt.LoadTypes.ELECTRICITY, unit=lt.Units.WATT,
                                                                               sankey_flow_direction=False)

        self.total_electricity_consumption_channel: cp.ComponentOutput = self.add_output(object_name=self.component_name,
                                                                               field_name=self.TotalElectricityConsumption,
                                                                               load_type=lt.LoadTypes.ELECTRICITY, unit=lt.Units.WATT,
                                                                               sankey_flow_direction=False)

        self.flexible_electricity: cp.ComponentOutput = self.add_output(object_name=self.component_name,
                                                                               field_name=self.FlexibleElectricity,
                                                                               load_type=lt.LoadTypes.ELECTRICITY, unit=lt.Units.WATT,
                                                                               sankey_flow_direction=False)


        self.building_temperature_modifier: cp.ComponentOutput = self.add_output(object_name=self.component_name,
                                                                               field_name=self.BuildingTemperatureModifier,
                                                                               load_type=lt.LoadTypes.TEMPERATURE, unit=lt.Units.CELSIUS,
                                                                               sankey_flow_direction=False)

        self.storage_temperature_modifier: cp.ComponentOutput = self.add_output(object_name=self.component_name,
                                                                                 field_name=self.StorageTemperatureModifier,
                                                                                 load_type=lt.LoadTypes.TEMPERATURE, unit=lt.Units.CELSIUS,
                                                                                 sankey_flow_direction=False)

        self.check_peak_shaving: cp.ComponentOutput = self.add_output(object_name=self.component_name, field_name=self.CheckPeakShaving,
                                                                      load_type=lt.LoadTypes.ANY, unit=lt.Units.ANY, sankey_flow_direction=False)

    def sort_source_weights_and_components(self) -> None:
        SourceTags = [elem.source_tags[0] for elem in self.my_component_inputs]
        SourceWeights = [elem.source_weight for elem in self.my_component_outputs]
        sortindex = sorted(range(len(SourceWeights)), key=lambda k: SourceWeights[k])
        self.source_weights_sorted = [SourceWeights[i] for i in sortindex]
        self.components_sorted = [SourceTags[i] for i in sortindex]

    def get_entries_of_type(self, component_type: lt.ComponentType) -> Any:
        return self.components_sorted.index(component_type)

    def build(self, mode: Any) -> None:
        self.mode = mode

    def write_to_report(self) -> None:
        pass

    def i_save_state(self) -> None:
        # abändern, siehe Storage
        pass  # self.previous_state = self.state

    def i_restore_state(self) -> None:
        pass  # self.state = self.previous_state

    def i_prepare_simulation(self) -> None:
        """ Prepares the simulation. """
        pass

    def i_doublecheck(self, timestep: int, stsv: cp.SingleTimeStepValues) -> None:
        pass

    def control_electricity_component_iterative(self, deltademand: float, stsv: cp.SingleTimeStepValues, weight_counter: int,
                                                component_type: lt.ComponentType) -> Any:
        is_battery = None
        # get previous signal and substract from total balance
        previous_signal = self.get_dynamic_input(stsv=stsv, tags=[component_type, lt.InandOutputType.ELECTRICITY_REAL], weight_counter=weight_counter)
        
        if component_type == lt.ComponentType.CAR_BATTERY:
            self.set_dynamic_output(stsv=stsv, tags=[component_type, lt.InandOutputType.ELECTRICITY_TARGET], weight_counter=weight_counter,
                                    output_value=deltademand)
            if previous_signal > 0:
                deltademand = deltademand - previous_signal

        # control from substracted balance
        elif component_type == lt.ComponentType.BATTERY:
            self.set_dynamic_output(stsv=stsv, tags=[component_type, lt.InandOutputType.ELECTRICITY_TARGET], weight_counter=weight_counter,
                                    output_value=deltademand)
            deltademand = deltademand - previous_signal

        elif component_type == lt.ComponentType.FUEL_CELL:
            if deltademand < 0:
                self.set_dynamic_output(stsv=stsv, tags=[component_type, lt.InandOutputType.ELECTRICITY_TARGET], weight_counter=weight_counter,
                                        output_value=-deltademand)
                deltademand = deltademand + previous_signal
                if deltademand > 0:
                    is_battery = self.get_entries_of_type(lt.ComponentType.BATTERY)
            else:
                self.set_dynamic_output(stsv=stsv, tags=[component_type, lt.InandOutputType.ELECTRICITY_TARGET], weight_counter=weight_counter,
                                        output_value=0)

        elif component_type in [lt.ComponentType.ELECTROLYZER, lt.ComponentType.HEAT_PUMP, lt.ComponentType.SMART_DEVICE]:

            if deltademand > 0:
                self.set_dynamic_output(stsv=stsv, tags=[component_type, lt.InandOutputType.ELECTRICITY_TARGET], weight_counter=weight_counter,
                                        output_value=deltademand)
                deltademand = deltademand - previous_signal
            else:
                self.set_dynamic_output(stsv=stsv, tags=[component_type, lt.InandOutputType.ELECTRICITY_TARGET], weight_counter=weight_counter,
                                        output_value=0)

        return deltademand, is_battery

    def postprocess_battery(self, deltademand: float, stsv: cp.SingleTimeStepValues, ind: int) -> Any:
        previous_signal = self.get_dynamic_input(stsv=stsv, tags=[self.components_sorted[ind]], weight_counter=self.source_weights_sorted[ind])
        self.set_dynamic_output(stsv=stsv, tags=[lt.ComponentType.BATTERY, lt.InandOutputType.ELECTRICITY_TARGET],
                                weight_counter=self.source_weights_sorted[ind], output_value=deltademand + previous_signal)
        return deltademand - previous_signal

    def optimize_own_consumption_iterative(self, delta_demand: float, stsv: cp.SingleTimeStepValues) -> None:
        skip_CHP = False
        for ind in range(len(self.source_weights_sorted)):
            component_type = self.components_sorted[ind]
            source_weight = self.source_weights_sorted[ind]
            if component_type in [lt.ComponentType.BATTERY, lt.ComponentType.FUEL_CELL, lt.ComponentType.ELECTROLYZER, lt.ComponentType.HEAT_PUMP,
                                  lt.ComponentType.SMART_DEVICE, lt.ComponentType.CAR_BATTERY]:
                if not skip_CHP or component_type in [lt.ComponentType.BATTERY, lt.ComponentType.ELECTROLYZER, lt.ComponentType.HEAT_PUMP,
                                                      lt.ComponentType.SMART_DEVICE, lt.ComponentType.CAR_BATTERY]:
                    delta_demand, is_battery = self.control_electricity_component_iterative(deltademand=delta_demand, stsv=stsv,
                                                                                            weight_counter=source_weight,
                                                                                            component_type=component_type)
                else:
                    self.set_dynamic_output(stsv=stsv, tags=[lt.ComponentType.FUEL_CELL, lt.InandOutputType.ELECTRICITY_TARGET],
                                            weight_counter=source_weight, output_value=0)
                if is_battery is not None:
                    delta_demand = self.postprocess_battery(deltademand=delta_demand, stsv=stsv, ind=is_battery)
                    skip_CHP = True

    def i_simulate(self, timestep: int, stsv: cp.SingleTimeStepValues, force_convergence: bool) -> None:
        if force_convergence:
            return

        if timestep == 0:
            self.sort_source_weights_and_components()

        ###ELECTRICITY#####

        # get production
        production = sum(self.get_dynamic_inputs(stsv=stsv, tags=[lt.InandOutputType.ELECTRICITY_PRODUCTION]))
        consumption_uncontrolled = sum(self.get_dynamic_inputs(stsv=stsv, tags=[lt.InandOutputType.ELECTRICITY_CONSUMPTION_UNCONTROLLED]))
        consumption_ems_controlled = sum(self.get_dynamic_inputs(stsv=stsv, tags=[lt.InandOutputType.ELECTRICITY_REAL]))

        # Production of Electricity positve sign
        # Consumption of Electricity negative sign
        flexible_electricity = production - consumption_uncontrolled
        electricity_to_grid = production - consumption_uncontrolled - consumption_ems_controlled
        if self.strategy == "optimize_own_consumption":
            self.optimize_own_consumption_iterative(delta_demand=electricity_to_grid, stsv=stsv)
            stsv.set_output_value(self.electricity_to_or_from_grid, electricity_to_grid)
            stsv.set_output_value(self.flexible_electricity, flexible_electricity)
        stsv.set_output_value(self.total_electricity_consumption_channel, consumption_uncontrolled + consumption_ems_controlled)
        if flexible_electricity > 0:
            stsv.set_output_value(self.building_temperature_modifier, 1)
            stsv.set_output_value(self.storage_temperature_modifier, 10)
        else:
            stsv.set_output_value(self.building_temperature_modifier, 0)
            stsv.set_output_value(self.storage_temperature_modifier, 0)
        '''
        elif self.strategy == "seasonal_storage":
            self.seasonal_storage(delta_demand=delta_demand, stsv=stsv)
        elif self.strategy == "peak_shaving_into_grid":
            self.peak_shaving_into_grid(delta_demand=delta_demand, limit_to_shave=limit_to_shave,stsv=stsv)
        elif self.strategy == "peak_shaving_from_grid":
            self.peak_shaving_from_grid(delta_demand=delta_demand, limit_to_shave=limit_to_shave,stsv=stsv)
        '''

        #######HEAT########
        # If comftortable temperature of building is to low heat with WarmWaterStorage the building
        # Solution with Control Signal Residence
        # not perfect solution!
        '''
        if self.temperature_residence<self.min_comfortable_temperature_residence:
            #heat
            #here has to be added how "strong" HeatingWater Storage can be discharged
            #Working with upper boarder?
        elif self.temperature_residence > self.max_comfortable_temperature_residence:
            #cool
        elif self.temperature_residence>self.min_comfortable_temperature_residence and self.temperature_residence<self.max_comfortable_temperature_residence:
        '''



""" Wraps components for use in the simulator. """
# clean
from typing import List, Dict, Any

import hisim.component as cp
import hisim.loadtypes as lt
from hisim import log


class ComponentWrapper:

    """ Wraps components for use. """

    def __init__(self, component: cp.Component, is_cachable: bool):
        """ Initializes the component wrapper.

        Used to handle the connection of inputs and outputs.
        """
        self.my_component: cp.Component = component
        self.component_inputs: List[cp.ComponentInput] = []
        self.component_outputs: List[cp.ComponentOutput] = []
        # self.cachedict: = {}
        self.is_cachable = is_cachable

    def clear(self):
        """ Clears properties to help with saving memory. """
        del self.my_component
        del self.component_inputs
        del self.component_outputs

    def register_component_outputs(self, all_outputs: List[cp.ComponentOutput]) -> None:
        """ Registers component outputs in the global list of components. """
        log.information("Registering component outputs on " + self.my_component.component_name)
        # register the output column
        output_columns = self.my_component.get_outputs()
        for col in output_columns:
            col.global_index = len(all_outputs)  # noqa
            for output in all_outputs:
                if output.full_name == col.full_name:
                    raise Exception("trying to register the same key twice: " + col.full_name)
            all_outputs.append(col)
            log.debug("Registered output " + col.full_name)
            self.component_outputs.append(col)

    def register_component_inputs(self, global_column_dict: Dict[str, Any]) -> None:
        """ Gets the inputs for the current component from the global column dict and puts them into component_inputs. """
        log.information("Registering component inputs for " + self.my_component.component_name)
        # look up input columns and cache, so we only have the correct columns saved
        input_columns: List[cp.ComponentInput] = self.my_component.get_input_definitions()
        for col in input_columns:
            global_column_entry = global_column_dict[col.fullname]
            self.component_inputs.append(global_column_entry)

    def save_state(self) -> None:
        """ Saves the state.

        This gets called at the beginning of a timestep and wraps the i_save_state
        i_save_state should always cache the current state at the beginning of a time step.
        """
        self.my_component.i_save_state()

    def doublecheck(self, timestep: int, stsv: cp.SingleTimeStepValues) -> None:
        """  Wrapper for i_doublecheck.

        Doublecheck is completely optional call that can be used while debugging to
        double check the component results after the iteration finished for a timestep.
        """
        self.my_component.i_doublecheck(timestep, stsv)

    def restore_state(self) -> None:
        """ Wrapper for i_restore_state.

        Gets called at the beginning of every iteration to return to the state at the beginning of the iteration.
        """
        self.my_component.i_restore_state()

    def calculate_component(self, timestep: int, stsv: cp.SingleTimeStepValues, force_convergence: bool) -> None:
        """ Wrapper for the core simulation function in each component. """
        self.my_component.i_simulate(timestep, stsv, force_convergence)

    def prepare_calculation(self):
        """ Wrapper for i_prepare_calculation. """
        log.information("Preparing " + self.my_component.component_name + " for simulation.")
        self.my_component.i_prepare_simulation()

    def connect_inputs(self, all_outputs: List[cp.ComponentOutput]) -> None:
        """ Connects cp.ComponentOutputs to ComponentInputs of WrapperComponent. """
        # Returns a List of ComponentInputs
        self.my_component.get_input_definitions()

        # Loop through lists of inputs of self component
        for cinput in self.my_component.inputs:
            # Adds to the ComponentInput List of ComponentWrapper
            self.component_inputs.append(cinput)

            # Creates a ComponentOutput variable
            global_output: cp.ComponentOutput

            # Loop through all the existent component outputs in the current simulation
            for global_output in all_outputs:
                # Check if ComponentOutput and ComponentInput match
                if global_output.component_name == cinput.src_object_name and global_output.field_name == cinput.src_field_name:
                    # Check if ComponentOutput and ComponentInput have the same units
                    if cinput.unit != global_output.unit:
                        # Check the use of "Units.Any"
                        if (cinput.unit == lt.Units.ANY and global_output.unit != lt.Units.ANY) or (
                                cinput.unit != lt.Units.ANY and global_output.unit == lt.Units.ANY):
                            log.warning(
                                f"The input {cinput.field_name} (cp: {cinput.component_name}, unit: {cinput.unit}) "
                                f"and output {global_output.field_name}(cp: {global_output.component_name}, unit: {global_output.unit}) "
                                f"might not have compatible units.")  #
                            # Connect, i.e, save ComponentOutput in ComponentInput
                            cinput.source_output = global_output
                            log.debug("Connected input '" + cinput.fullname + "' to '" + global_output.full_name + "'")
                        else:
                            raise SystemError(
                                f"The input {cinput.field_name} (cp: {cinput.component_name}, unit: {cinput.unit}) and "
                                f"output {global_output.field_name}(cp: {global_output.component_name}, unit: {global_output.unit}) "
                                f"do not have the same unit!")  #
                    else:
                        # Connect, i.e, save ComponentOutput in ComponentInput
                        cinput.source_output = global_output
                        log.debug(f"connected input {cinput.fullname} to {global_output.full_name}")

            # Check if there are inputs that have been not connected
            if cinput.is_mandatory and cinput.source_output is None:
                raise SystemError(
                    f"The ComponentInput {cinput.field_name} (cp: {cinput.component_name}, "
                    f"unit: {cinput.unit}) is not connected to any ComponentOutput.")  #



# Owned
from hisim.component import Component, SingleTimeStepValues, ComponentInput, ComponentOutput
from hisim.simulationparameters import SimulationParameters
from hisim import loadtypes as lt
from dataclasses import dataclass
from dataclasses_json import dataclass_json
from typing import List
__authors__ = "Frank Burkrad, Maximilian Hillen"
__copyright__ = "Copyright 2021, the House Infrastructure Project"
__credits__ = ["Noah Pflugradt"]
__license__ = ""
__version__ = ""
__maintainer__ = "Maximilian Hillen"
__email__ = "maximilian.hillen@rwth-aachen.de"
__status__ = ""

@dataclass_json
@dataclass
class GenericGasHeaterConfig:
    """
    Gas Heater Config
    """
    is_modulating : bool
    P_th_min : float           # [W]
    P_th_max : float           # [W]
    eff_th_min : float         # [-]
    eff_th_max : float         # [-]
    delta_T : float            # [°C]
    mass_flow_max :   float  # kg/s ## -> ~0.07 P_th_max / (4180 * delta_T)
    temperature_max :  float   # [°C]
    temperaturedelta : float   # [°C]
    power_max :float           # [W]

class GasHeater(Component):
    """
    Gets Control Signal and calculates on base of it Massflow and Temperature of Massflow
    """
    #Input
    ControlSignal = "ControlSignal" # at which Procentage is the GasHeater modulating [0..1]
    MassflowInputTemperature = "MassflowInputTemperature"

    #Output
    MassflowOutput= "Hot Water Energy Output"
    MassflowOutputTemperature= "MassflowOutputTemperature"
    GasDemand = "GasDemand"
    ThermalOutputPower="ThermalOutputPower"

    def __init__(self, my_simulation_parameters: SimulationParameters, config : GenericGasHeaterConfig) -> None:
        super().__init__(name="GasHeater", my_simulation_parameters=my_simulation_parameters)
        self.control_signal: ComponentInput = self.add_input(self.component_name, GasHeater.ControlSignal, lt.LoadTypes.ANY, lt.Units.PERCENT, True)
        self.mass_inp_temp: ComponentInput = self.add_input(self.component_name, GasHeater.MassflowInputTemperature, lt.LoadTypes.WATER, lt.Units.CELSIUS, True)


        self.mass_out: ComponentOutput = self.add_output(self.component_name, GasHeater.MassflowOutput, lt.LoadTypes.WATER, lt.Units.KG_PER_SEC)
        self.mass_out_temp: ComponentOutput = self.add_output(self.component_name, GasHeater.MassflowOutputTemperature, lt.LoadTypes.WATER, lt.Units.CELSIUS)
        self.gas_demand: ComponentOutput = self.add_output(self.component_name, GasHeater.GasDemand, lt.LoadTypes.GAS, lt.Units.KWH)
        self.p_th: ComponentOutput = self.add_output(object_name=self.component_name,
                                                     field_name=self.ThermalOutputPower,
                                                     load_type=lt.LoadTypes.HEATING,
                                                     unit=lt.Units.WATT)

        self.P_th_min = config.P_th_min
        self.P_th_max = config.power_max
        self.eff_th_min = config.eff_th_min
        self.eff_th_max = config.eff_th_max
        self.temperature_max = config.temperature_max
        self.temperaturedelta = config.temperaturedelta

    @staticmethod
    def get_default_config():
        config=GenericGasHeaterConfig(
                    temperaturedelta = 10,
                    power_max = 12_000,
                    is_modulating = True,
                    P_th_min = 1_000  ,# [W]
                    P_th_max = 12_000 , # [W]
                    eff_th_min = 0.60  ,# [-]
                    eff_th_max = 0.90  ,# [-]
                    delta_T = 25,
                    mass_flow_max = 12_000 / (4180 * 25),  # kg/s ## -> ~0.07 P_th_max / (4180 * delta_T)
                    temperature_max = 80  # [°C])
                    )
        return config

    def i_prepare_simulation(self) -> None:
        """ Prepares the simulation. """
        pass

    def write_to_report(self) -> List[str]:
        """Writes a report."""
        lines: List = []
        return lines

    def i_save_state(self) -> None:
        pass

    def i_restore_state(self) -> None:
        pass

    def i_doublecheck(self, timestep: int, stsv: SingleTimeStepValues) -> None:
        pass

    def i_simulate(self, timestep: int, stsv: SingleTimeStepValues, force_convergence: bool) -> None:
        control_signal = stsv.get_input_value(self.control_signal)
        if control_signal > 1:
            raise Exception("Expected a control signal between 0 and 1")
        if control_signal < 0:
            raise Exception("Expected a control signal between 0 and 1")

        #Calculate Eff
        d_eff_th = (self.eff_th_max - self.eff_th_min)

        if control_signal*self.P_th_max<self.P_th_min:
            maximum_power=self.P_th_min
            eff_th_real=self.eff_th_min
        else:
            maximum_power=control_signal*self.P_th_max
            eff_th_real = self.eff_th_min + d_eff_th * control_signal

        gas_power = maximum_power *eff_th_real* control_signal
        cw = 4182
        mass_out_temp=self.temperaturedelta+stsv.get_input_value(self.mass_inp_temp)
        mass_out=gas_power/(cw*self.temperaturedelta)
        p_th=cw*mass_out*(mass_out_temp-stsv.get_input_value(self.mass_inp_temp))

        stsv.set_output_value(self.p_th, gas_power)  # efficiency
        stsv.set_output_value(self.mass_out_temp, mass_out_temp)  # efficiency
        stsv.set_output_value(self.mass_out, mass_out)  # efficiency
        stsv.set_output_value(self.gas_demand, gas_power)  # gas consumption



# -*- coding: utf-8 -*-
"""
Created on Tue Aug  9 09:06:13 2022

@author: m.alfouly
"""
#import numpy as np

# Owned
#from hisim.components.weather import Weather
from hisim import component as cp
from hisim.loadtypes import LoadTypes, Units
from hisim.simulationparameters import SimulationParameters
from hisim.components.building import Building
#from hisim.components.weather import Weather
#from hisim.components.loadprofilegenerator_connector import Occupancy
from hisim import log


class PIDState:
    """ Represents the current internal state of the PID. """

    def __init__(self,
                 integrator: float,
                 derivator: float):
        """ Initializes the state of the PID. """
        self.Integrator: float = integrator
        self.Derivator: float = derivator

    def clone(self):
        return PIDState(self.Integrator, self.Derivator)


class PIDController(cp.Component):
    # Inputs
    TemperatureMean = "Residence Temperature"  # uncontrolled temperature

    # ouput 
    ElectricityOutputPID = "ElectricityOutputPID"
    ThermalEnergyDelivered = "ThermalEnergyDelivered"
    error_pvalue = "error_p_value"
    error_dvalue = "error_d_value"
    error_ivalue = "error_i_value"
    error = "error_value"
    derivator = "derivator"
    integrator = "integrator"

    def __init__(self, my_simulation_parameters: SimulationParameters, ki: float = 1, kp: float = 1, kd: float =1) -> None:
        super().__init__("PIDController", my_simulation_parameters=my_simulation_parameters)
        self.build()
        # --------------------------------------------------
        # control saturation
        self.MV_min = 0
        self.MV_max = 5000
        self.ki = ki
        self.kp = kp
        self.kd = kd
        self.state = PIDState(integrator=0, derivator=0)
        self.previous_state = self.state.clone()

        self.t_mC: cp.ComponentInput = self.add_input(self.component_name,
                                                      self.TemperatureMean,
                                                      LoadTypes.TEMPERATURE,
                                                      Units.CELSIUS,
                                                      True)

        self.electric_power: cp.ComponentOutput = self.add_output(self.component_name,
                                                                  self.ElectricityOutputPID,
                                                                  LoadTypes.ELECTRICITY,
                                                                  Units.WATT)
        self.error_pvalue_output: cp.ComponentOutput = self.add_output(self.component_name,
                                                                       self.error_pvalue,
                                                                       LoadTypes.ELECTRICITY,
                                                                       Units.WATT)
        self.error_ivalue_output: cp.ComponentOutput = self.add_output(self.component_name,
                                                                       self.error_ivalue,
                                                                       LoadTypes.ELECTRICITY,
                                                                       Units.WATT)
        self.error_dvalue_output: cp.ComponentOutput = self.add_output(self.component_name,
                                                                       self.error_dvalue,
                                                                       LoadTypes.ELECTRICITY,
                                                                       Units.WATT)
        self.error_output: cp.ComponentOutput = self.add_output(self.component_name,
                                                                self.error,
                                                                LoadTypes.ELECTRICITY,
                                                                Units.WATT)

        self.derivator_output: cp.ComponentOutput = self.add_output(self.component_name,
                                                                self.derivator,
                                                                LoadTypes.ELECTRICITY,
                                                                Units.WATT)
        self.integrator_output: cp.ComponentOutput = self.add_output(self.component_name,
                                                             self.integrator,
                                                             LoadTypes.ELECTRICITY,
                                                             Units.WATT)

    def get_building_default_connections(self):
        log.information("setting building default connections in Heatpumpcontroller")
        connections = []
        building_classname = Building.get_classname()
        connections.append(
            cp.ComponentConnection(PIDController.TemperatureMean, building_classname, Building.TemperatureMean))
        # connections.append(cp.ComponentConnection(PIDController.SolarGainThroughWindows, building_classname,                                                  Building.SolarGainThroughWindows))

        return connections

    # def get_occupancy_default_connections(self):
    #     log.information("setting occupancy default connections")
    #     connections = []
    #     occupancy_classname = Occupancy.get_classname()
    #     connections.append(
    #         cp.ComponentConnection(PIDController.HeatingByResidents, occupancy_classname, Occupancy.HeatingByResidents))
    #     return connections
    #
    # def get_weather_default_connections(self):
    #     log.information("setting weather default connections")
    #     connections = []
    #     weather_classname = Weather.get_classname()
    #     connections.append(
    #         cp.ComponentConnection(PIDController.TemperatureOutside, weather_classname, Weather.TemperatureOutside))
    #     return connections

    def build(self):
        """ For calculating internal things and preparing the simulation. """
        # Sth
        pass

    def i_save_state(self):
        """ Saves the internal state at the beginning of each timestep. """
        self.previous_state = self.state.clone()

    def i_restore_state(self):
        """ Restores the internal state after each iteration. """
        self.state = self.previous_state.clone()

    def i_doublecheck(self, timestep: int, stsv: cp.SingleTimeStepValues) -> None:
        """ Double check results after iteration. """
        pass

    def write_to_report(self):
        """ Logs the most important config stuff to the report. """
        lines = []
        lines.append("PID Controller")
        lines.append("Control algorithm of the Air conditioner is: PI \n")
        # lines.append(f"Controller Proportional gain is {self.Kp:4.2f} \n")
        # lines.append(f"Controller Integral gain is {self.Ki:4.2f} \n")
        return lines

    def i_simulate(self, timestep: int, stsv: cp.SingleTimeStepValues, force_convergence: bool) -> None:
        """ Core smulation function. """
        if force_convergence:
            return
        # Retrieve building temperature
        building_temperature_t_mc = stsv.get_input_value(self.t_mC)
        set_point: float = 22.0

        tau_r: float = 1.0
        tauI: float = tau_r * 1863
        Kplant: float = 1.8750 / 1000  # 0.15773831/(10000) #
        Kp: float = self.kp  # 2500   # (tau_r * Kplant)  # controller Proportional gain
        Ki: float = self.ki  # 0.2 #  Kp / tauI  # integral gain
        Kd: float = self.kd  #  200

        error = set_point - building_temperature_t_mc  # e(tk)
        p_value = Kp * error
        d_value = Kd * (error - self.state.Derivator)
        self.state.Derivator = error
        self.state.Integrator = self.state.Integrator + error

        limit = 5000
        if self.state.Integrator > limit:
            self.state.Integrator = limit
        elif self.state.Integrator < -limit:
            self.state.Integrator = -limit

        i_value = self.state.Integrator * Ki
        manipulated_variable = p_value + i_value + d_value

        stsv.set_output_value(self.error_pvalue_output, p_value)
        stsv.set_output_value(self.error_dvalue_output, d_value)
        stsv.set_output_value(self.error_ivalue_output, i_value)
        stsv.set_output_value(self.error_output, error)
        stsv.set_output_value(self.integrator_output, self.state.Integrator)
        stsv.set_output_value(self.derivator_output, self.state.Derivator)
        stsv.set_output_value(self.electric_power, manipulated_variable)

        # update state for next iteration
        #self.state.previous_value = building_temperature_t_mc

    def determine_conditions(self, current_temperature: float, set_point: float) -> str:
        """ For determining heating and cooling mode and implementing a dead zone. Currently disabled. """
        offset_in_degree_celsius = 0.5
        maximum_heating_set_temp = set_point + offset_in_degree_celsius
        minimum_cooling_set_temp = set_point - offset_in_degree_celsius

        mode = 'off'
        if mode == 'heating' and current_temperature > maximum_heating_set_temp:  # 23 22.5
            return 'off'
        if mode == 'cooling' and current_temperature < minimum_cooling_set_temp:  # 24 21.5
            mode = 'off'
        if mode == 'off':
            if current_temperature < set_point:  # 21
                mode = 'heating'
            if current_temperature > set_point:  # 26
                mode = 'cooling'
        return mode



# Generic/Built-in
import datetime
import math
import os
from dataclasses import dataclass
from functools import lru_cache
from typing import Optional, Any

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import pvlib
from dataclasses_json import dataclass_json

# Owned
from hisim import component as cp
from hisim import loadtypes as lt
from hisim import log
from hisim import sim_repository
from hisim import utils
from hisim.component import ConfigBase
from hisim.components.weather import Weather
from hisim.simulationparameters import SimulationParameters

__authors__ = "Vitor Hugo Bellotto Zago"
__copyright__ = "Copyright 2021, the House Infrastructure Project"
__credits__ = ["Noah Pflugradt"]
__license__ = "MIT"
__version__ = "0.1"
__maintainer__ = "Vitor Hugo Bellotto Zago"
__email__ = "vitor.zago@rwth-aachen.de"
__status__ = "development"

"""
The functions cited in this module are at some degree based on the tsib project:

[tsib-kotzur]:
Kotzur, Leander, Detlef Stolten, and Hermann-Josef Wagner. Future grid load of the residential building sector. No. RWTH-2018-231872. Lehrstuhl für Brennstoffzellen (FZ Jülich), 2019.
ID: http://hdl.handle.net/2128/21115
    http://nbn-resolving.org/resolver?verb=redirect&identifier=urn:nbn:de:0001-2019020614
    
The implementation of the tsib project can be found under the following repository:
https://github.com/FZJ-IEK3-VSA/tsib
"""

temp_model = pvlib.temperature.TEMPERATURE_MODEL_PARAMETERS["sapm"]["open_rack_glass_glass"]


@lru_cache(maxsize=16)
def simPhotovoltaicFast(dni_extra: Any = None, DNI: Any = None, DHI: Any = None, GHI: Any = None, azimuth: Any = None, apparent_zenith: Any = None,
        temperature: Any = None, wind_speed: Any = None, surface_azimuth: float = 180, surface_tilt: float = 30) -> Any:
    """
    Simulates a defined PV array with the Sandia PV Array Performance Model.
    The implementation is done in accordance with following tutorial:
    https://github.com/pvlib/pvlib-python/blob/master/docs/tutorials/tmy_to_power.ipynb

    Parameters
    ----------
    surface_tilt: int or float, optional (default:30)
        Tilt angle of of the array in degree.
    surface_azimuth: int or float, optional (default:180)
        Azimuth angle of of the array in degree. 180 degree means south,
        90 degree east and 270 west.
    losses: float, optional (default: 0.1)
        Losses due to soiling, mismatch, diode connections, dc wiring etc.
    Returns
    --------
    """

    poa_irrad = pvlib.irradiance.get_total_irradiance(surface_tilt, surface_azimuth, apparent_zenith, azimuth, DNI, GHI, DHI, dni_extra)

    pvtemps = pvlib.temperature.sapm_cell(poa_irrad["poa_global"], temperature, wind_speed, **temp_model)

    pv_dc = pvlib.pvsystem.pvwatts_dc(poa_irrad["poa_global"], temp_cell=pvtemps, pdc0=1, gamma_pdc=-0.002, temp_ref=25.0)
    if math.isnan(pv_dc):
        pv_dc = 0
    return pv_dc


def simPhotovoltaicSimple(dni_extra=None, DNI=None, DHI=None, GHI=None, azimuth=None, apparent_zenith=None, temperature=None, wind_speed=None,
        surface_tilt=30, surface_azimuth=180, albedo=0.2):
    """
    Simulates a defined PV array with the Sandia PV Array Performance Model.
    The implementation is done in accordance with following tutorial:
    https://github.com/pvlib/pvlib-python/blob/master/docs/tutorials/tmy_to_power.ipynb

    Based on the tsib project @[tsib-kotzur] (Check header)

    Parameters
    ----------
    tmy_data: pandas.DataFrame(), required
        Weatherfile in the format of a tmy file.
    surface_tilt: int or float, optional (default:30)
        Tilt angle of of the array in degree.
    surface_azimuth: int or float, optional (default:180)
        Azimuth angle of of the array in degree. 180 degree means south,
        90 degree east and 270 west.
    albedo: float, optional (default: 0.2)
        Reflection coefficient of the surrounding area.
    losses: float, optional (default: 0.1)
        Losses due to soiling, mismatch, diode connections, dc wiring etc.
    load_module_data: Boolean, optional (default: False)
        If True the module data base is loaded from the Sandia Website.
        Otherwise it is loaded from this relative path
            '\\profiles\\PV-Modules\\sandia_modules.csv'.
    module_name: str, optional (default:'Hanwha_HSL60P6_PA_4_250T__2013_')
        Module name. The string must be existens in Sandia Module database.
    integrateInverter: bool, optional (default: True)
        If an inverter shall be added to the simulation, providing the photovoltaic output after the inverter.
    inverter_name: str, optional (default: 'ABB__MICRO_0_25_I_OUTD_US_208_208V__CEC_2014_')
        Type of inverter.

    Returns
    --------
    """
    # automatic pd time series in future pvlib version
    # calculate airmass
    airmass = pvlib.atmosphere.get_relative_airmass(apparent_zenith)
    # use perez model to calculate the plane of array diffuse sky radiation
    poa_sky_diffuse = pvlib.irradiance.perez(surface_tilt, surface_azimuth, DHI, np.float64(DNI), dni_extra, apparent_zenith, azimuth, airmass, )
    # calculate ground diffuse with specified albedo
    poa_ground_diffuse = pvlib.irradiance.get_ground_diffuse(surface_tilt, GHI, albedo=albedo)
    # calculate angle of incidence
    aoi = pvlib.irradiance.aoi(surface_tilt, surface_azimuth, apparent_zenith, azimuth)
    # calculate plane of array irradiance
    poa_irrad = pvlib.irradiance.poa_components(aoi, np.float64(DNI), poa_sky_diffuse, poa_ground_diffuse)
    # calculate pv cell and module temperature
    temp_model = pvlib.temperature.TEMPERATURE_MODEL_PARAMETERS["sapm"]["open_rack_glass_glass"]
    pvtemps = pvlib.temperature.sapm_cell(poa_irrad["poa_global"], temperature, wind_speed, **temp_model)

    pv_dc = pvlib.pvsystem.pvwatts_dc(poa_irrad["poa_global"], temp_cell=pvtemps, pdc0=1, gamma_pdc=-0.002, temp_ref=25.0)
    if math.isnan(pv_dc):
        pv_dc = 0
    return pv_dc


@dataclass_json
@dataclass
class PVSystemConfig(ConfigBase):
    @classmethod
    def get_main_classname(cls):
        """ Returns the full class name of the base class. """
        return PVSystem.get_full_classname()

    # parameter_string: str
    # my_simulation_parameters: SimulationParameters
    name: str
    time: int
    location: str
    module_name: str
    integrate_inverter: bool
    inverter_name: str
    power: float
    azimuth: float
    tilt: float
    load_module_data: bool
    source_weight: int
    @classmethod
    def get_default_PV_system(cls):
        """ Gets a default PV system. """
        return PVSystemConfig(time=2019, power=10E3, load_module_data=False, module_name="Hanwha_HSL60P6_PA_4_250T__2013_", integrate_inverter=True,
            inverter_name="ABB__MICRO_0_25_I_OUTD_US_208_208V__CEC_2014_", name='PVSystem', azimuth=180, tilt=30, source_weight=0, location="Aachen")


class PVSystem(cp.Component):
    """
    Simulates PV Output based on weather data and peak power.

    Parameters:
    -----------------------------------------------------
    time : int, optional
        Simulation timeline. The default is 2019.
    location : str, optional
        Object Location with temperature and solar data. The default is "Aachen".
    power : float, optional
        Power in kWp to be provided by the PV System. The default is 10E3.
    load_module_data : bool
        Access the PV data base (True) or not (False). The default is False
    module_name : str, optional
        The default is "Hanwha_HSL60P6_PA_4_250T__2013_"
    integrate_inverter, bool, optional
        Consider inverter efficiency in the calculation (True) or not (False). The default is True.
    inverter_name : str, optional
        The default is "ABB__MICRO_0_25_I_OUTD_US_208_208V__CEC_2014_".
    azimuth : float, optional
        Panel azimuth from north in °. The default is 180°.
    tilt : float, optional
        Panel tilt from horizontal. The default is 90°.
    source_weight : int, optional
        Weight of component, relevant if there is more than one PV System, defines hierachy in control. The default is 1.
    name : str, optional
        Name of pv panel within simulation. The default is 'PVSystem'

    """
    # Inputs
    TemperatureOutside = "TemperatureOutside"
    DirectNormalIrradiance = "DirectNormalIrradiance"
    DirectNormalIrradianceExtra = "DirectNormalIrradianceExtra"
    DiffuseHorizontalIrradiance = "DiffuseHorizontalIrradiance"
    GlobalHorizontalIrradiance = "GlobalHorizontalIrradiance"
    Azimuth = "Azimuth"
    ApparentZenith = "ApparentZenith"
    WindSpeed = "WindSpeed"

    # Outputs
    ElectricityOutput = "ElectricityOutput"

    # Similar components to connect to:
    # 1. Weather
    @utils.measure_execution_time
    def __init__(self, my_simulation_parameters: SimulationParameters, config: PVSystemConfig,
                 my_simulation_repository: Optional[sim_repository.SimRepository] = None) -> None:
        self.my_simulation_parameters = my_simulation_parameters
        self.pvconfig = config
        self.data: Any
        super().__init__(self.pvconfig.name + '_w' + str(self.pvconfig.source_weight), my_simulation_parameters=my_simulation_parameters)

        self.t_outC: cp.ComponentInput = self.add_input(self.component_name, self.TemperatureOutside, lt.LoadTypes.TEMPERATURE, lt.Units.CELSIUS,
                                                        True)

        self.DNIC: cp.ComponentInput = self.add_input(self.component_name, self.DirectNormalIrradiance, lt.LoadTypes.IRRADIANCE,
                                                      lt.Units.WATT_PER_SQUARE_METER, True)

        self.DNIextraC: cp.ComponentInput = self.add_input(self.component_name, self.DirectNormalIrradianceExtra, lt.LoadTypes.IRRADIANCE,
                                                           lt.Units.WATT_PER_SQUARE_METER, True)

        self.DHIC: cp.ComponentInput = self.add_input(self.component_name, self.DiffuseHorizontalIrradiance, lt.LoadTypes.IRRADIANCE,
                                                      lt.Units.WATT_PER_SQUARE_METER, True)

        self.GHIC: cp.ComponentInput = self.add_input(self.component_name, self.GlobalHorizontalIrradiance, lt.LoadTypes.IRRADIANCE,
                                                      lt.Units.WATT_PER_SQUARE_METER, True)

        self.azimuthC: cp.ComponentInput = self.add_input(self.component_name, self.Azimuth, lt.LoadTypes.ANY, lt.Units.DEGREES, True)

        self.apparent_zenithC: cp.ComponentInput = self.add_input(self.component_name, self.ApparentZenith, lt.LoadTypes.ANY, lt.Units.DEGREES, True)

        self.wind_speedC: cp.ComponentInput = self.add_input(self.component_name, self.WindSpeed, lt.LoadTypes.SPEED, lt.Units.METER_PER_SECOND, True)

        self.electricity_outputC: cp.ComponentOutput = self.add_output(object_name=self.component_name, field_name=PVSystem.ElectricityOutput,
                                                                       load_type=lt.LoadTypes.ELECTRICITY, unit=lt.Units.WATT, postprocessing_flag=[lt.InandOutputType.ELECTRICITY_PRODUCTION])
        self.add_default_connections(Weather, self.get_weather_default_connections())

    @staticmethod
    def get_default_config(power: float = 10E3, source_weight: int = 1) -> Any:
        config = PVSystemConfig(name='PVSystem', time=2019, location="Aachen", module_name="Hanwha_HSL60P6_PA_4_250T__2013_", integrate_inverter=True,
            inverter_name="ABB__MICRO_0_25_I_OUTD_US_208_208V__CEC_2014_", power=power, azimuth=180, tilt=30, load_module_data=False,
            source_weight=source_weight)
        return config

    def get_weather_default_connections(self):
        log.information("setting weather default connections")
        connections = []
        weather_classname = Weather.get_classname()
        connections.append(cp.ComponentConnection(PVSystem.TemperatureOutside, weather_classname, Weather.TemperatureOutside))
        connections.append(cp.ComponentConnection(PVSystem.DirectNormalIrradiance, weather_classname, Weather.DirectNormalIrradiance))
        connections.append(cp.ComponentConnection(PVSystem.DirectNormalIrradianceExtra, weather_classname, Weather.DirectNormalIrradianceExtra))
        connections.append(cp.ComponentConnection(PVSystem.DiffuseHorizontalIrradiance, weather_classname, Weather.DiffuseHorizontalIrradiance))
        connections.append(cp.ComponentConnection(PVSystem.GlobalHorizontalIrradiance, weather_classname, Weather.GlobalHorizontalIrradiance))
        connections.append(cp.ComponentConnection(PVSystem.Azimuth, weather_classname, Weather.Azimuth))
        connections.append(cp.ComponentConnection(PVSystem.ApparentZenith, weather_classname, Weather.ApparentZenith))
        connections.append(cp.ComponentConnection(PVSystem.WindSpeed, weather_classname, Weather.WindSpeed))
        return connections

    def i_restore_state(self):
        pass

    def write_to_report(self):
        lines = []
        lines.append("Name: {}".format(self.component_name))
        lines.append("Power: {:3.0f} kWp".format(self.pvconfig.power * 1E-3))
        lines.append("Module: {}".format(self.pvconfig.module_name))
        lines.append("Inverter: {}".format(self.pvconfig.inverter_name))
        return lines

    def i_simulate(self, timestep: int, stsv: cp.SingleTimeStepValues, force_convergence: bool) -> None:

        if hasattr(self, "output"):
            # if(len(self.output) < timestep)
            #   raise Exception("Somehow the precalculated list of values for the PV system seems to be incorrect. Please delete the cache.")
            stsv.set_output_value(self.electricity_outputC, self.output[timestep] * self.pvconfig.power)
        else:
            DNI = stsv.get_input_value(self.DNIC)
            dni_extra = stsv.get_input_value(self.DNIextraC)
            DHI = stsv.get_input_value(self.DHIC)
            GHI = stsv.get_input_value(self.GHIC)
            azimuth = stsv.get_input_value(self.azimuthC)
            temperature = stsv.get_input_value(self.t_outC)
            wind_speed = stsv.get_input_value(self.wind_speedC)
            apparent_zenith = stsv.get_input_value(self.apparent_zenithC)

            # ac_power = self.simPhotovoltaic2(dni_extra=dni_extra,
            #                                 DNI=DNI,
            #                                 DHI=DHI,
            #                                 GHI=GHI,
            #                                 azimuth=azimuth,
            #                                 apparent_zenith=apparent_zenith,
            #                                 temperature=temperature,
            #                                 wind_speed=wind_speed)
            # ac_power = simPhotovoltaicSimple(
            #    dni_extra=dni_extra,
            #                                 DNI=DNI,
            #                                 DHI=DHI,
            #                                 GHI=GHI,
            #                                 azimuth=azimuth,
            #                                 apparent_zenith=apparent_zenith,
            #                                 temperature=temperature,
            #                                 wind_speed=wind_speed)

            ac_power = simPhotovoltaicFast(dni_extra=dni_extra, DNI=DNI, DHI=DHI, GHI=GHI, azimuth=azimuth, apparent_zenith=apparent_zenith,
                temperature=temperature, wind_speed=wind_speed, surface_azimuth=self.pvconfig.azimuth, surface_tilt=self.pvconfig.tilt)

            resultingvalue = ac_power * self.pvconfig.power
            # if you wanted to access the temperature forecast from the weather component:
            # val = self.simulation_repository.get_entry(Weather.Weather_Temperature_Forecast_24h)
            stsv.set_output_value(self.electricity_outputC, resultingvalue)
            self.data[timestep] = ac_power
            if timestep + 1 == self.data_length:
                database = pd.DataFrame(self.data, columns=["output"])

                database.to_csv(self.cache_filepath, sep=",", decimal=".", index=False)

        if self.my_simulation_parameters.system_config.predictive == True and self.my_simulation_parameters.system_config.prediction_horizon:
            last_forecast_timestep = int(
                timestep + self.my_simulation_parameters.system_config.prediction_horizon / self.my_simulation_parameters.seconds_per_timestep)
            if (last_forecast_timestep > len(self.output)):
                last_forecast_timestep = len(self.output)
            pvforecast = [self.output[t] * self.pvconfig.power for t in range(timestep, last_forecast_timestep)]
            self.simulation_repository.set_dynamic_entry(component_type=lt.ComponentType.PV, source_weight=self.pvconfig.source_weight, entry=pvforecast)

            if timestep == 1:
                # delete weather data for PV preprocessing from dictionary -> save memory
                if self.simulation_repository.exist_entry(Weather.Weather_DirectNormalIrradianceExtra_yearly_forecast):
                    self.simulation_repository.delete_entry(Weather.Weather_DirectNormalIrradianceExtra_yearly_forecast)
                    self.simulation_repository.delete_entry(Weather.Weather_DirectNormalIrradiance_yearly_forecast)
                    self.simulation_repository.delete_entry(Weather.Weather_DiffuseHorizontalIrradiance_yearly_forecast)
                    self.simulation_repository.delete_entry(Weather.Weather_GlobalHorizontalIrradiance_yearly_forecast)
                    self.simulation_repository.delete_entry(Weather.Weather_Azimuth_yearly_forecast)
                    self.simulation_repository.delete_entry(Weather.Weather_ApparentZenith_yearly_forecast)
                    self.simulation_repository.delete_entry(Weather.Weather_TemperatureOutside_yearly_forecast)
                    self.simulation_repository.delete_entry(Weather.Weather_WindSpeed_yearly_forecast)


    def i_save_state(self) -> None:
        pass

    def i_doublecheck(self, timestep: int, stsv: cp.SingleTimeStepValues) -> None:
        pass

    def i_prepare_simulation(self) -> None:
        """ Prepares the component for the simulation """
        log.information(self.pvconfig.to_json())  # type: ignore
        file_exists, self.cache_filepath = utils.get_cache_file("PVSystem", self.pvconfig, self.my_simulation_parameters)

        if file_exists:
            self.output = pd.read_csv(self.cache_filepath, sep=',', decimal='.')['output'].tolist()
            if len(self.output) != self.my_simulation_parameters.timesteps:
                raise Exception("Reading the cached PV values seems to have failed. Expected " + str(
                    self.my_simulation_parameters.timesteps) + " values, but got " + str(len(self.output)))
        else:
            self.coordinates = self.simulation_repository.get_entry("weather_location")
            # Factor to guarantee peak power based on module with 250 Wh
            self.ac_power_factor = math.ceil((self.pvconfig.power * 1e3) / 250)

            # when predictive control is activated, the PV simulation is run beforhand to make forecasting easier
            if self.my_simulation_parameters.system_config.predictive:
                # get yearly weather data from dictionary
                dni_extra = self.simulation_repository.get_entry(Weather.Weather_DirectNormalIrradianceExtra_yearly_forecast)
                DNI = self.simulation_repository.get_entry(Weather.Weather_DirectNormalIrradiance_yearly_forecast)
                DHI = self.simulation_repository.get_entry(Weather.Weather_DiffuseHorizontalIrradiance_yearly_forecast)
                GHI = self.simulation_repository.get_entry(Weather.Weather_GlobalHorizontalIrradiance_yearly_forecast)
                azimuth = self.simulation_repository.get_entry(Weather.Weather_Azimuth_yearly_forecast)
                apparent_zenith = self.simulation_repository.get_entry(Weather.Weather_ApparentZenith_yearly_forecast)
                temperature = self.simulation_repository.get_entry(Weather.Weather_TemperatureOutside_yearly_forecast)
                wind_speed = self.simulation_repository.get_entry(Weather.Weather_WindSpeed_yearly_forecast)

                x = []
                for i in range(self.my_simulation_parameters.timesteps):
                    x.append(simPhotovoltaicFast(dni_extra[i], DNI[i], DHI[i], GHI[i], azimuth[i], apparent_zenith[i], temperature[i], wind_speed[i],
                                                 self.pvconfig.azimuth, self.pvconfig.tilt))

                self.output = x

                database = pd.DataFrame(self.output, columns=["output"])

                database.to_csv(self.cache_filepath, sep=",", decimal=".", index=False)

            else:
                self.data = [0] * self.my_simulation_parameters.timesteps
                self.data_length = self.my_simulation_parameters.timesteps

        self.modules = pd.read_csv(os.path.join(utils.HISIMPATH["photovoltaic"]["modules"]), index_col=0, )

        self.inverters = pd.read_csv(os.path.join(utils.HISIMPATH["photovoltaic"]["inverters"]), index_col=0, )

        self.temp_model = pvlib.temperature.TEMPERATURE_MODEL_PARAMETERS["sapm"]["open_rack_glass_glass"]

        # load the sandia data
        if self.pvconfig.load_module_data:
            # load module data online
            modules = pvlib.pvsystem.retrieve_sam(name="SandiaMod")
            self.module = modules[self.pvconfig.module_name]
            # get inverter data
            inverters = pvlib.pvsystem.retrieve_sam("cecinverter")
            self.inverter = inverters[self.pvconfig.inverter_name]
        else:
            # load module and inverter data from csv
            module = self.modules[self.pvconfig.module_name]
            self.module = pd.to_numeric(module, errors="coerce")

            inverter = self.inverters[self.pvconfig.inverter_name]
            self.inverter = pd.to_numeric(inverter,                                          errors="coerce")
            # self.power = self.power  # self.module_name =  module_name  # self.inverter_name = inverter_name  # self.integrateInverter = integrateInverter  # self.simPhotovoltaicSimpleJit = simPhotovoltaicSimple

    # def plot(self) -> None:
    #     # Plots ac_power. One day is represented by 1440 steps.
    #     # self.ac_power.iloc[0:7200].plot()
    #     plt.plot(self.data)
    #     plt.ylabel("Power [W]")
    #     plt.xlabel("Time")
    #     plt.show()

    def interpolate(self, pd_database: Any, year: Any) -> Any:
        firstday = pd.Series([0.0], index=[pd.to_datetime(datetime.datetime(year - 1, 12, 31, 23, 0), utc=True).tz_convert("Europe/Berlin")])
        lastday = pd.Series(pd_database[-1], index=[pd.to_datetime(datetime.datetime(year, 12, 31, 22, 59), utc=True).tz_convert("Europe/Berlin")])
        # pd_database = pd_database.append(firstday)
        pd_database = pd_database.append(lastday)
        pd_database = pd_database.sort_index()
        return pd_database.resample('1T').asfreq().interpolate(method='linear').tolist()

    def simPhotovoltaic2(self, dni_extra=None, DNI=None, DHI=None, GHI=None, azimuth=None, apparent_zenith=None, temperature=None, wind_speed=None,
            surface_tilt=30, surface_azimuth=180, albedo=0.2):
        """
        Simulates a defined PV array with the Sandia PV Array Performance Model.
        The implementation is done in accordance with following tutorial:
        https://github.com/pvlib/pvlib-python/blob/master/docs/tutorials/tmy_to_power.ipynb

        Based on the tsib project @[tsib-kotzur] (Check header)

        Parameters
        ----------
        tmy_data: pandas.DataFrame(), required
            Weatherfile in the format of a tmy file.
        surface_tilt: int or float, optional (default:30)
            Tilt angle of of the array in degree.
        surface_azimuth: int or float, optional (default:180)
            Azimuth angle of of the array in degree. 180 degree means south,
            90 degree east and 270 west.
        albedo: float, optional (default: 0.2)
            Reflection coefficient of the surrounding area.
        losses: float, optional (default: 0.1)
            Losses due to soiling, mismatch, diode connections, dc wiring etc.
        load_module_data: Boolean, optional (default: False)
            If True the module data base is loaded from the Sandia Website.
            Otherwise it is loaded from this relative path
                '\\profiles\\PV-Modules\\sandia_modules.csv'.
        module_name: str, optional (default:'Hanwha_HSL60P6_PA_4_250T__2013_')
            Module name. The string must be existens in Sandia Module database.
        integrateInverter: bool, optional (default: True)
            If an inverter shall be added to the simulation, providing the photovoltaic output after the inverter.
        inverter_name: str, optional (default: 'ABB__MICRO_0_25_I_OUTD_US_208_208V__CEC_2014_')
            Type of inverter.

        Returns
        --------
        """
        # automatic pd time series in future pvlib version
        # calculate airmass
        airmass = pvlib.atmosphere.get_relative_airmass(apparent_zenith)
        # use perez model to calculate the plane of array diffuse sky radiation
        poa_sky_diffuse = pvlib.irradiance.perez(surface_tilt, surface_azimuth, DHI, np.float64(DNI), dni_extra, apparent_zenith, azimuth, airmass, )
        # calculate ground diffuse with specified albedo
        poa_ground_diffuse = pvlib.irradiance.get_ground_diffuse(surface_tilt, GHI, albedo=albedo)
        # calculate angle of incidence
        aoi = pvlib.irradiance.aoi(surface_tilt, surface_azimuth, apparent_zenith, azimuth)
        # calculate plane of array irradiance
        poa_irrad = pvlib.irradiance.poa_components(aoi, np.float64(DNI), poa_sky_diffuse, poa_ground_diffuse)
        # calculate pv cell and module temperature
        # temp_model = pvlib.temperature.TEMPERATURE_MODEL_PARAMETERS["sapm"]["open_rack_glass_glass"]
        pvtemps = pvlib.temperature.sapm_cell(poa_irrad["poa_global"], temperature, wind_speed, **self.temp_model)

        # calculate effective irradiance on pv module
        sapm_irr = pvlib.pvsystem.sapm_effective_irradiance(module=self.module, poa_direct=poa_irrad["poa_direct"],
            poa_diffuse=poa_irrad["poa_diffuse"], airmass_absolute=airmass, aoi=aoi, )
        # calculate pv performance
        sapm_out = pvlib.pvsystem.sapm(sapm_irr, module=self.module, temp_cell=pvtemps, )
        # calculate peak load of single module [W]
        peak_load = self.module.loc["Impo"] * self.module.loc["Vmpo"]
        ac_power = pd.DataFrame()
        
        if self.pvconfig.integrate_inverter:
            # calculate load after inverter
            iv_load = pvlib.inverter.sandia(inverter=self.inverter, v_dc=sapm_out["v_mp"], p_dc=sapm_out["p_mp"])
            ac_power = iv_load / peak_load
        else:
            # load in [kW/kWp]
            ac_power = sapm_out["p_mp"] / peak_load

        if math.isnan(ac_power):  # type: ignore
            raise ValueError("AC power was NAN")
            # ac_power = 0.0

        # ac_power = ac_power * self.time_correction_factor
        # ac_power = ac_power

        # data = [DHI,
        #        DNI,
        #        GHI,
        #        dni_extra,
        #        aoi,
        #        apparent_zenith,
        #        azimuth,
        #        airmass,
        #        wind_speed]
        # if timestep % 60 == 0 and timestep < 1442:
        #    log.information(data)
        #    log.information("Timestep:{} , AcPower: {}".format(timestep, ac_power))

        return ac_power

#
# def readTRY(location="Aachen", year=2010):
#     """
#     Reads a test reference year file and gets the GHI, DHI and DNI from it.
#
#     Based on the tsib project @[tsib-kotzur] (Check header)
#
#     Parameters
#     -------
#     try_num: int (default: 4)
#         The region number of the test reference year.
#     year: int (default: 2010)
#         The year. Only data for 2010 and 2030 available
#     """
#     # get the correct file path
#     filepath = os.path.join(utils.HISIMPATH["weather"][location])
#
#     # get the geoposition
#     with open(filepath + ".dat", encoding="utf-8") as fp:
#         lines = fp.readlines()
#         location_name = lines[0].split(maxsplit=2)[2].replace('\n', '')
#         lat = float(lines[1][20:37])
#         lon = float(lines[2][15:30])
#     location = {"name": location_name, "latitude": lat, "longitude": lon}
#
#     # check if time series data already exists as .csv with DNI
#     if os.path.isfile(filepath + ".csv"):
#         data = pd.read_csv(filepath + ".csv", index_col=0, parse_dates=True, sep=";", decimal=",")
#         data.index = pd.to_datetime(data.index, utc=True).tz_convert("Europe/Berlin")
#     # else read from .dat and calculate DNI etc.
#     # else:
#         # get data
#         data = pd.read_csv(filepath + ".dat", sep=r"\s+", skiprows=([i for i in range(0, 31)]))
#         data.index = pd.date_range("{}-01-01 00:00:00".format(year), periods=8760, freq="H", tz="Europe/Berlin")
#         data["GHI"] = data["D"] + data["B"]
#         data = data.rename(columns={"D": "DHI", "t": "T", "WG": "WS"})
#
#         # calculate direct normal
#         data["DNI"] = calculateDNI(data["B"], lon, lat)  # data["DNI"] = data["B"]
#
#         # save as .csv  # data.to_csv(filepath + ".csv",sep=";",decimal=",")
#     return data, location

#
# def calculateDNI(directHI, lon, lat, zenith_tol=87.0):
#     """
#     Calculates the direct NORMAL irradiance from the direct horizontal irradiance with the help of the PV lib.
#
#     Based on the tsib project @[tsib-kotzur] (Check header)
#
#     Parameters
#     ----------
#     directHI: pd.Series with time index
#         Direct horizontal irradiance
#     lon: float
#         Longitude of the location
#     lat: float
#         Latitude of the location
#     zenith_tol: float, optional
#         Avoid cosines of values above a certain zenith angle of in order to avoid division by zero.
#
#     Returns
#     -------
#     DNI: pd.Series
#     """
#     solarPos = pvlib.solarposition.get_solarposition(directHI.index, lat, lon)
#     solarPos["apparent_zenith"][solarPos.apparent_zenith > zenith_tol] = zenith_tol
#     DNI = directHI.div(solarPos["apparent_zenith"].apply(math.radians).apply(math.cos))
#     DNI = DNI.fillna(0)
#     if DNI.isnull().values.any():
#         raise ValueError("Something went wrong...")
#     return DNI



""" Is used to create all necessary json files. """
from typing import List
from dataclasses import dataclass, field, asdict
import json
from dataclasses_json import dataclass_json
from economic_parameters import EconomicParameters
import hisim.loadtypes as lt


@dataclass_json
@dataclass()
class ComponentCost:

    """ Defines the investment parameters of components. """

    component: lt.ComponentType = lt.ComponentType.BATTERY
    capacity_unit: lt.Units = lt.Units.WATT_HOUR
    cost_unit: lt.Units = lt.Units.EURO
    co2_unit: lt.Units = lt.Units.KG
    time_unit: lt.Units = lt.Units.YEARS
    costfactor_unit: lt.Units = lt.Units.PERCENT
    capacity_for_cost: List[float] = field(default_factory=list)
    cost_per_capacity: List[float] = field(default_factory=list)
    time: List[float] = field(default_factory=list)
    costfactor_per_time: List[float] = field(default_factory=list)
    capacity_for_co2: List[float] = field(default_factory=list)
    co2_per_capacity: List[float] = field(default_factory=list)


def create_componentcost_file(
        component: lt.ComponentType, capacity_unit: lt.Units, capacity_for_cost: List,
        cost_per_capacity: List, time: List, costfactor_per_time: List, capacity_for_co2: List,
        co2_per_capacity: List) -> None:
    """Component Cost file is created."""

    costfile = ComponentCost(
        component=component, capacity_unit=capacity_unit, capacity_for_cost=capacity_for_cost,
        cost_per_capacity=cost_per_capacity, time=time, costfactor_per_time=costfactor_per_time,
        capacity_for_co2=capacity_for_co2, co2_per_capacity=co2_per_capacity)
    costfile_written = json.dumps(asdict(costfile))

    with open('ComponentCost' + component.value + '.json', 'w', encoding="utf-8") as outfile:
        outfile.write(costfile_written)


def write_batterycost_file():
    """Battery Cost is written to json file."""
    create_componentcost_file(
        component=lt.ComponentType.BATTERY, capacity_unit=lt.Units.WATT_HOUR, capacity_for_cost=[500, 1000, 2000, 10000],
        cost_per_capacity=[1000, 1800, 2500, 5000], time=[2022, 2030, 2050], costfactor_per_time=[100, 90, 60],
        capacity_for_co2=[1000, 10000], co2_per_capacity=[200, 2000])


def write_heatpump_cost_file():
    """Heatpump Cost is written to json file."""
    create_componentcost_file(
        component=lt.ComponentType.HEAT_PUMP, capacity_unit=lt.Units.WATT_HOUR, capacity_for_cost=[500, 1000, 2000, 10000],
        cost_per_capacity=[1000, 1800, 2500, 5000], time=[2022, 2030, 2050], costfactor_per_time=[100, 90, 60],
        capacity_for_co2=[1000, 10000], co2_per_capacity=[200, 2000])


def write_chp_cost_file():
    """Chp Cost is written to json file."""
    create_componentcost_file(
        component=lt.ComponentType.PREDICTIVE_CONTROLLER, capacity_unit=lt.Units.WATT_HOUR, capacity_for_cost=[500, 1000, 2000, 10000],
        cost_per_capacity=[1000, 1800, 2500, 5000], time=[2022, 2030, 2050], costfactor_per_time=[100, 90, 60],
        capacity_for_co2=[1000, 10000], co2_per_capacity=[200, 2000])


def create_economicparameters_file(
        insulation_bought: bool,
        insulation_threshold: float,
        pv_bought: bool,
        pv_threshold: float,
        smart_devices_bought: bool,
        smart_devices_threshold: float,
        heatpump_bought: bool,
        heatpump_threshold: float,
        buffer_bought: bool,
        buffer_threshold: float,
        battery_bought: bool,
        battery_threshold: float,
        h2system_bought: bool,
        h2system_threshold: float,
        chp_bought: bool,
        chp_threshold: float,
        electrolyzer_bought: bool,
        electrolyzer_threshold: float,
        surpluscontroller_bought: bool,
        surpluscontroller_threshold: float,
        ev_bought: bool,
        ev_threshold: float) -> None:
    """Economic Parameters are written to json file."""

    economic_parameters_file = EconomicParameters(
        insulation_bought=insulation_bought,
        insulation_threshold=insulation_threshold,
        pv_bought=pv_bought,
        pv_threshold=pv_threshold,
        smart_devices_bought=smart_devices_bought,
        smart_devices_threshold=smart_devices_threshold,
        heatpump_bought=heatpump_bought,
        heatpump_threshold=heatpump_threshold,
        buffer_bought=buffer_bought,
        buffer_threshold=buffer_threshold,
        battery_bought=battery_bought,
        battery_threshold=battery_threshold,
        h2system_bought=h2system_bought,
        h2system_threshold=h2system_threshold,
        chp_bought=chp_bought,
        chp_threshold=chp_threshold,
        electrolyzer_bought=electrolyzer_bought,
        electrolyzer_threshold=electrolyzer_threshold,
        surpluscontroller_bought=surpluscontroller_bought,
        surpluscontroller_threshold=surpluscontroller_threshold,
        ev_bought=ev_bought,
        ev_threshold=ev_threshold)
    economic_parameters_file = json.dumps(asdict(economic_parameters_file))

    with open('EconomicParameters.json', 'w', encoding="utf-8") as outfile:
        outfile.write(economic_parameters_file)


def write_economicparameters_file():
    """Economic Parameters are written to a json file."""
    create_economicparameters_file(
        insulation_bought=True,
        insulation_threshold=1e3,
        pv_bought=True,
        pv_threshold=1e3,
        smart_devices_bought=True,
        smart_devices_threshold=5e2,
        heatpump_bought=False,
        heatpump_threshold=1e4,
        buffer_bought=True,
        buffer_threshold=1e2,
        battery_bought=True,
        battery_threshold=1e3,
        h2system_bought=False,
        h2system_threshold=5e3,
        chp_bought=False,
        chp_threshold=5e3,
        electrolyzer_bought=False,
        electrolyzer_threshold=5e3,
        surpluscontroller_bought=False,
        surpluscontroller_threshold=5e3,
        ev_bought=True,
        ev_threshold=2e4)



# Import packages from standard library or the environment e.g. pandas, numpy etc.
from copy import deepcopy
from dataclasses import dataclass
from dataclasses_json import dataclass_json
from hplib import hplib as hpl

# Import modules from HiSim
from hisim.component import Component, ComponentInput, ComponentOutput, SingleTimeStepValues
from hisim.loadtypes import LoadTypes, Units
from hisim.simulationparameters import SimulationParameters
from typing import Optional
__authors__ = "Tjarko Tjaden, Hauke Hoops, Kai Rösken"
__copyright__ = "Copyright 2021, the House Infrastructure Project"
__credits__ = "..."
__license__ = "MIT"
__version__ = "0.1"
__maintainer__ = "Tjarko Tjaden"
__email__ = "tjarko.tjaden@hs-emden-leer.de"
__status__ = "development"

@dataclass_json
@dataclass
class HeatPumpHplibConfig:
    model: str
    group_id: int
    t_in: float
    t_out_val: float
    p_th_set: float

class HeatPumpHplib(Component):
    """
    Simulate heat pump efficiency (cop) as well as electrical (p_el) & 
    thermal power (p_th), massflow (m_dot) and output temperature (t_out).
    Relevant simulation parameters are loaded within the init for a 
    specific or generic heat pump type.
    """

    # Inputs
    OnOffSwitch = "OnOffSwitch"                                 # 1 = on, 0 = 0ff
    TemperatureInputPrimary = "TemperatureInputPrimary"         # °C
    TemperatureInputSecondary = "TemperatureInputSecondary"     # °C
    TemperatureAmbient = "TemperatureAmbient"                   # °C

    # Outputs
    ThermalOutputPower = "ThermalOutputPower"                   # W
    ElectricalInputPower = "ElectricalInputPower"               # W
    COP = "COP"                                                 # -
    TemperatureOutput = "TemperatureOutput"                     # °C
    MassFlowOutput = "MassFlowOutput"                           # kg/s
    TimeOn = "TimeOn"                                           # s
    TimeOff = "TimeOff"                                         # s

    def __init__(self, my_simulation_parameters: SimulationParameters, config: HeatPumpHplibConfig):
        """
        Loads the parameters of the specified heat pump.

        Parameters
        ----------
        model : str
            Name of the heat pump model or "Generic".
        group_id : numeric, default 0
            only for model "Generic": Group ID for subtype of heat pump. [1-6].
        t_in : numeric, default 0
            only for model "Generic": Input temperature :math:`T` at primary side of the heat pump. [°C]
        t_out_val : numeric, default 0
            only for model "Generic": Output temperature :math:`T` at secondary side of the heat pump. [°C]
        p_th_set : numeric, default 0
            only for model "Generic": Thermal output power at setpoint t_in, t_out. [W]

        Returns
        ----------
        parameters : pd.DataFrame
            Data frame containing the model parameters.
        """
        super().__init__(name="HeatPump", my_simulation_parameters=my_simulation_parameters)

        self.model = config.model

        self.group_id = config.group_id

        self.t_in = config.t_in

        self.t_out_val = config.t_out_val

        self.p_th_set = config.p_th_set

        # Component has states
        self.state = HeatPumpState()
        self.previous_state = deepcopy(self.state)

        # Load parameters from heat pump database
        self.parameters = hpl.get_parameters(self.model, self.group_id,
                                             self.t_in, self.t_out_val, self.p_th_set)

        # Define component inputs
        self.on_off_switch: ComponentInput = self.add_input(object_name=self.component_name,
                                                            field_name=self.OnOffSwitch,
                                                            load_type=LoadTypes.ANY,
                                                            unit=Units.ANY,
                                                            mandatory=True)

        self.t_in_primary: ComponentInput = self.add_input(object_name=self.component_name,
                                                           field_name=self.TemperatureInputPrimary,
                                                           load_type=LoadTypes.TEMPERATURE,
                                                           unit=Units.CELSIUS,
                                                           mandatory=True)
        
        self.t_in_secondary: ComponentInput = self.add_input(object_name=self.component_name,
                                                             field_name=self.TemperatureInputSecondary,
                                                             load_type=LoadTypes.TEMPERATURE,
                                                             unit=Units.CELSIUS,
                                                             mandatory=True)
        
        self.t_amb: ComponentInput = self.add_input(object_name=self.component_name,
                                                    field_name=self.TemperatureAmbient,
                                                    load_type=LoadTypes.TEMPERATURE,
                                                    unit=Units.CELSIUS,
                                                    mandatory=True)

        # Define component outputs
        self.p_th: ComponentOutput = self.add_output(object_name=self.component_name,
                                                     field_name=self.ThermalOutputPower,
                                                     load_type=LoadTypes.HEATING,
                                                     unit=Units.WATT)
        
        self.p_el: ComponentOutput = self.add_output(object_name=self.component_name,
                                                     field_name=self.ElectricalInputPower,
                                                     load_type=LoadTypes.ELECTRICITY,
                                                     unit=Units.WATT)

        self.cop: ComponentOutput = self.add_output(object_name=self.component_name,
                                                    field_name=self.COP,
                                                    load_type=LoadTypes.ANY,
                                                    unit=Units.ANY)

        self.t_out: ComponentOutput = self.add_output(object_name=self.component_name,
                                                      field_name=self.TemperatureOutput,
                                                      load_type=LoadTypes.HEATING,
                                                      unit=Units.CELSIUS)

        self.m_dot: ComponentOutput = self.add_output(object_name=self.component_name,
                                                      field_name=self.MassFlowOutput,
                                                      load_type=LoadTypes.VOLUME,
                                                      unit=Units.KG_PER_SEC)
        
        self.time_on: ComponentOutput = self.add_output(object_name=self.component_name,
                                                        field_name=self.TimeOn,
                                                        load_type=LoadTypes.TIME,
                                                        unit=Units.SECONDS)
        
        self.time_off: ComponentOutput = self.add_output(object_name=self.component_name,
                                                         field_name=self.TimeOff,
                                                         load_type=LoadTypes.TIME,
                                                         unit=Units.SECONDS)
    @staticmethod
    def get_defaul_config():
        config= HeatPumpHplibConfig(model="Generic",
                                    group_id= -1,
                                    t_in= -300,
                                    t_out_val= -300,
                                    p_th_set= -30)
        return config
    def i_save_state(self) -> None:
        self.previous_state = deepcopy(self.state)

    def i_restore_state(self) -> None:
        self.state = deepcopy(self.previous_state)

    def i_doublecheck(self, timestep: int,  stsv: SingleTimeStepValues) -> None:
        pass

    def i_simulate(self, timestep: int, stsv: SingleTimeStepValues,  force_convergence: bool) -> None:
        # Parameter
        time_on_min = 600 # [s]
        time_off_min = time_on_min
        
        # Load input values
        on_off:float = stsv.get_input_value(self.on_off_switch)
        t_in_primary = stsv.get_input_value(self.t_in_primary)
        t_in_secondary = stsv.get_input_value(self.t_in_secondary)
        t_amb = stsv.get_input_value(self.t_amb)
        time_on = self.state.time_on
        time_off = self.state.time_off
        on_off_previous = self.state.on_off_previous
        
        # Overwrite on_off to realize minimum time of or time off
        if on_off_previous == 1 & time_on < time_on_min:
            on_off=1
        elif on_off_previous == 0 & time_off < time_off_min:
            on_off=0

        # OnOffSwitch
        if on_off == 1:
            # Calulate outputs
            results = hpl.simulate(t_in_primary, t_in_secondary, self.parameters, t_amb)
            p_th=results['P_th'].values[0]
            p_el=results['P_el'].values[0]
            cop=results['COP'].values[0]
            t_out=results['T_out'].values[0]
            m_dot=results['m_dot'].values[0]
            time_on = time_on + self.my_simulation_parameters.seconds_per_timestep
            time_off = 0
        else:
            # Calulate outputs
            p_th = 0
            p_el = 0
            cop = None
            t_out = None
            m_dot = 0
            time_off = time_off +  self.my_simulation_parameters.seconds_per_timestep
            time_on = 0
        
        # write values for output time series
        stsv.set_output_value(self.p_th, p_th)
        stsv.set_output_value(self.p_el, p_el)
        stsv.set_output_value(self.cop, cop)
        stsv.set_output_value(self.t_out, t_out)
        stsv.set_output_value(self.m_dot, m_dot)
        stsv.set_output_value(self.time_on, time_on)
        stsv.set_output_value(self.time_off, time_off)

        # write values to state
        self.state.time_on = time_on
        self.state.time_off = time_off
        self.state.on_off_previous = on_off

@dataclass
class HeatPumpState:
    time_on: int = 0
    time_off: int = 0
    on_off_previous: float = 0
    



# Generic/Built-in
import pandas as pd
import json
import numpy as np
import math as ma
from typing import Optional, List
from os import path

# Owned
from hisim import component as cp
from hisim import loadtypes as lt
from hisim import utils
from hisim.components import generic_pv_system
from hisim.components import generic_price_signal
from hisim.simulationparameters import SimulationParameters
from hisim.components.configuration import HouseholdWarmWaterDemandConfig
from hisim.components.configuration import PhysicsConfig

__authors__ = "Johanna Ganglbauer"
__copyright__ = "Copyright 2021, the House Infrastructure Project"
__credits__ = ["Noah Pflugradt"]
__license__ = "MIT"
__version__ = "0.1"
__maintainer__ = "Vitor Hugo Bellotto Zago"
__email__ = "vitor.zago@rwth-aachen.de"
__status__ = "development"

class SmartDeviceState:
    """Component representing smart appliance

    Parameters:
    -----------
    actual power : float, optional
        Power of smart appliance at given timestep
    timestep_of_activation : int, optional
        Timestep, where the device was activated. The default is -999.
    time_to_go : integer, optional
        Duration of the power profile, which follows for the nex time steps. The default is 0.
    position : integer, optional
        Index of demand profile relevent for the given timestep. The default is 0.
    """

    def __init__( self, actual_power : float = 0, timestep_of_activation : int = -999, time_to_go : int = 0, position : int = 0 ):
        self.actual_power = actual_power
        self.timestep_of_activation = timestep_of_activation
        self.time_to_go = time_to_go
        self.position = position

    def clone( self ):
        return SmartDeviceState( self.actual_power, self.timestep_of_activation, self.time_to_go, self.position )

    def run( self, timestep : int, electricity_profile : List[float] )-> None:

        #device activation
        if timestep > self.timestep_of_activation + self.time_to_go:
            self.timestep_of_activation = timestep
            self.time_to_go = len( electricity_profile )
            self.actual_power = electricity_profile[ 0 ]

        if timestep < self.timestep_of_activation + self.time_to_go:
            #device is running
            self.actual_power = electricity_profile[ timestep - self.timestep_of_activation ]

        #device deactivation
        if timestep == self.timestep_of_activation + self.time_to_go:
            self.position = self.position + 1
            self.time_to_go = 0
            self.actual_power = 0

class SmartDevice( cp.Component ):
    """
    Class component that provides availablity and profiles of flexible smart devices like shiftable (in time) washing machines and dishwashers.
    Data provided or based on LPG exports.

    Parameters
    -----------------------------------------------
    profile: string
        profile code corresponded to the family or residents configuration
    """

    #mandatory Inputs
    l3_DeviceActivation = "l3_DeviceActivation"

    # mandatory Outputs
    ElectricityOutput = "ElectricityOutput"

    #optional Inputs
    ElectricityTarget = "ElectricityTarget"

    def __init__( self,
                  identifier: str,
                  source_weight: int,
                  my_simulation_parameters: SimulationParameters,
                  smart_devices_included: bool):
        super().__init__ ( name = identifier.replace("/", "-") + '_w' + str(source_weight), my_simulation_parameters = my_simulation_parameters )

        self.build( identifier = identifier, source_weight = source_weight, seconds_per_timestep = my_simulation_parameters.seconds_per_timestep )

        if my_simulation_parameters.system_config.clever and smart_devices_included:
            postprocessing_flag = [lt.InandOutputType.ELECTRICITY_CONSUMPTION_EMS_CONTROLLED]
        else:
            postprocessing_flag = [lt.InandOutputType.ELECTRICITY_CONSUMPTION_UNCONTROLLED]

        #mandatory Output
        self.ElectricityOutputC: cp.ComponentOutput = self.add_output(
            object_name=self.component_name, field_name=self.ElectricityOutput, load_type=lt.LoadTypes.ELECTRICITY,
            unit=lt.Units.WATT, postprocessing_flag=postprocessing_flag)

        self.ElectricityTargetC: cp.ComponentInput = self.add_input(
            object_name=self.component_name, field_name=self.ElectricityTarget,
            load_type=lt.LoadTypes.ELECTRICITY, unit=lt.Units.WATT, mandatory=False )

    def i_save_state( self ) -> None:
        self.previous_state : SmartDeviceState = self.state.clone( )

    def i_restore_state(self)  -> None:
        self.state : SmartDeviceState = self.previous_state.clone( )

    def i_doublecheck(self, timestep: int, stsv: cp.SingleTimeStepValues)  -> None:
        pass

    def i_prepare_simulation(self) -> None:
        """ Prepares the simulation. """
        pass

    def i_simulate(self, timestep: int, stsv: cp.SingleTimeStepValues,  force_conversion: bool )  -> None:

        #initialize power
        self.state.actual_power = 0

        # if not already running: check if activation makes sense
        if timestep > self.state.timestep_of_activation + self.state.time_to_go:
            if timestep > self.earliest_start[self.state.position]: # can be turnod on
                #initialize next activation
                activation:float = timestep + 10
                #if surplus controller is connected get related signal
                if self.ElectricityTargetC.source_output is not None:
                    electricity_target = stsv.get_input_value(self.ElectricityTargetC)
                    if electricity_target >= self.electricity_profile[self.state.position][0]:
                        activation = timestep
                #if last possible switch on force activation
                if timestep >= self.latest_start[ self.state.position ]: #needs to be activated
                    activation = timestep

                if timestep == activation:
                    self.state.run( timestep, self.electricity_profile[ self.state.position ] )

        #run device if it was already activated
        else:
            self.state.run( timestep, self.electricity_profile[ self.state.position ] )

        stsv.set_output_value( self.ElectricityOutputC, self.state.actual_power )

    def build( self, identifier: str, source_weight: int, seconds_per_timestep : int = 60 ) -> None:

        #load smart device profile
        smart_device_profile = [ ]
        filepath = path.join(utils.HISIMPATH["utsp_reports"], "FlexibilityEvents.HH1.json")
        f = open( filepath )
        smart_device_profile = json.load( f )
        f.close()

        if not smart_device_profile:
            raise NameError( 'LPG data for smart appliances is missing or located missleadingly' )

        #initializing relevant data
        earliest_start, latest_start, electricity_profile = [ ], [ ], [ ]

        minutes_per_timestep = seconds_per_timestep / 60

        if not minutes_per_timestep.is_integer( ):
            raise TypeError( 'Up to now smart appliances have only been implemented for time resolutions corresponding to multiples of one minute' )
        minutes_per_timestep = int( minutes_per_timestep )

        #reading in data from json file and adopting to given time resolution
        for sample in smart_device_profile :
            device_name = str( sample[ 'Device' ][ 'Name' ] )
            if device_name == identifier:
                #earliest start in given time resolution -> integer value
                x = sample[ 'EarliestStart' ][ 'ExternalStep' ]
                #skip if occurs in calibration days (negative sign )
                if x < 0:
                    continue
                # timestep (in minutes) the profile is shifted in the first step of the external time resolution
                offset = minutes_per_timestep -  x % minutes_per_timestep
                #earliest start in given time resolution -> float value
                x = x / minutes_per_timestep
                #latest start in given time resolution
                y = sample[ 'LatestStart' ][ 'ExternalStep' ] /  minutes_per_timestep
                #number of timesteps in given time resolution -> integer value
                z = ma.ceil( x + sample[ 'TotalDuration' ] / minutes_per_timestep ) - ma.floor( x )
                #earliest and latest start in new time resolution -> integer value
                earliest_start.append( ma.floor( x ) )
                latest_start.append( ma.ceil( y ) )

                #get shiftable load profile
                el = sample[ 'Profiles' ][ 2 ][ 'TimeOffsetInSteps' ] * [ 0 ] + sample[ 'Profiles' ][ 2 ][ 'Values' ]

                #average profiles given in 1 minute resolution to given time resolution
                elem_el = [ ]
                #append first timestep which may not fill  the entire 15 minutes
                elem_el.append( sum( el[ : offset ] ) / offset )

                for i in range( z - 2 ):
                    elem_el.append( sum( el[ offset + minutes_per_timestep * i : offset + ( i + 1 ) * minutes_per_timestep ] ) / minutes_per_timestep )

                last = el[ offset + ( i + 1 ) * minutes_per_timestep : ]
                if offset != minutes_per_timestep:
                    elem_el.append( sum( last ) / ( minutes_per_timestep - offset ) )
                electricity_profile.append( elem_el )

        self.source_weight = source_weight
        self.earliest_start = earliest_start + [ self.my_simulation_parameters.timesteps ] #append value to continue simulation after last necesary run of flexible device at end of year
        self.latest_start = latest_start + [ self.my_simulation_parameters.timesteps + 999 ] #append value to continue simulation after last necesary run of smart device at end of year
        self.electricity_profile = electricity_profile
        self.state = SmartDeviceState( )
        self.previous_state = SmartDeviceState( )

    def write_to_report(self) -> List[str]:
        lines: List[str] = []
        lines.append("DeviceName: {}".format(self.component_name))
        return lines



from hisim.component import Component, SingleTimeStepValues, ComponentInput, ComponentOutput
from hisim import loadtypes as lt
import copy
from hisim.components.configuration import PhysicsConfig
from hisim import utils
#from math import pi
#from math import floor
from hisim.simulationparameters import SimulationParameters
import hisim.log as log
import pandas as pd
import os
from dataclasses import dataclass
from dataclasses_json import dataclass_json
import math
from typing import Any, List
__authors__ = "Frank Burkrad, Maximilian Hillen,"
__copyright__ = "Copyright 2021, the House Infrastructure Project"
__credits__ = ["Noah Pflugradt"]
__license__ = ""
__version__ = ""
__maintainer__ = "Maximilian Hillen"
__email__ = "maximilian.hillen@rwth-aachen.de"
__status__ = "development"

@dataclass_json
@dataclass
class CHPConfig:
    """
    CHP Config
    """
    name:str
    min_operation_time : float
    min_idle_time :float
    gas_type : str
    operating_mode :str
    p_el_max :float       # [W]
    is_modulating :bool
    P_el_min :float       # [W]
    P_th_min :float       # [W]
    eff_el_min :float     # [-]
    eff_th_min :float     # [-]
    mass_flow_max:float   #kg/s
    P_el_max :float       # [W]
    P_th_max :float       # [W]
    eff_el_max :float     # [-]
    eff_th_max :float     # [-]
    temperature_max:float
    delta_T:float

class CHPConfigAdvanced:

    # system_name = "BlueGEN15"
    # system_name = "Dachs 0.8"
    # system_name = "Test_KWK"
    # system_name = "Dachs G2.9"
    # system_name = "HOMER"
    system_name = "BlueGen BG15"

    df = pd.read_excel(os.path.join(utils.HISIMPATH["chp_system"], 'mock_up_efficiencies.xlsx'), index_col=0)

    df_specific = df.loc[str(system_name)]

    if str(df_specific['is_modulating']) == 'Yes':
        is_modulating = True
        P_el_min = df_specific['P_el_min']
        P_th_min = df_specific['P_th_min']
        P_total_min = df_specific['P_total_min']
        eff_el_min = df_specific['eff_el_min']
        eff_th_min = df_specific['eff_th_min']

    elif str(df_specific['is_modulating']) == 'No':
        is_modulating = False
    else:
        log.error("Modulation is not defined. Modulation must be 'Yes' or 'No'")
        raise ValueError

    P_el_max = df_specific['P_el_max']
    P_th_max = df_specific['P_th_max']
    P_total_max = df_specific['P_total_max']        # maximum fuel consumption
    eff_el_max = df_specific['eff_el_max']
    eff_th_max = df_specific['eff_th_max']
    mass_flow_max = df_specific['mass_flow (dT=20°C)']
    temperature_max = df_specific['temperature_max']
    delta_T=10

class CHPState:
    def __init__(self,
                 start_timestep=None,
                 electricity_output = 0.0,
                 cycle_number=None):
        self.start_timestep=start_timestep
        self.electricity_output = electricity_output
        self.cycle_number = cycle_number
        if self.electricity_output == 0.0:
            self.activation = 0
        elif self.electricity_output > 0.0:
            self.activation = 1
        else:
            raise Exception("Impossible CHPState.")
class CHP(Component):
    """
    Simulate chp efficiency (cop) as well as electrical (p_el) &
    thermal power (p_th), massflow (m_dot) and output temperature (t_out).
    """
    # Inputs
    ControlSignal = "ControlSignal" # at which Procentage is the CHP modulating [0..1]
    MassflowInputTemperature = "MassflowInputTemperature"
    ElectricityFromCHPTarget="ElectricityFromCHPTarget"
    HydrogenNotReleased="HydrogenNotReleased"
    #OperatingModelSignal="OperatingModelSignal" #-->Wärme oder Stromgeführt. Nötig?

    #Output
    MassflowOutput= "Hot Water Energy Output"
    MassflowOutputTemperature= "MassflowOutputTemperature"
    ElectricityOutput="ElectricityOutput"
    GasDemandTarget = "GasDemandTarget"
    NumberofCycles = "NumberofCycles"
    ThermalOutputPower = "ThermalOutputPower"
    GasDemandReal="GasDemandReal"
    def __init__(self, my_simulation_parameters: SimulationParameters, config: CHPConfig)  -> None:
        super().__init__(name=config.name, my_simulation_parameters=my_simulation_parameters)
        self.min_operation_time = config.min_operation_time
        self.min_idle_time = config.min_idle_time
        self. gas_type = config.gas_type #Gas Type can be "Hydrogen" or "Methan"
        self.operating_mode=config.operating_mode #operating_mode=["both","heat","electricity"]

        self.number_of_cycles = 0
        self.number_of_cycles_previous = copy.deepcopy(self.number_of_cycles)
        self.state = CHPState(start_timestep=int(0),cycle_number=0)
        self.previous_state = copy.deepcopy(self.state)

        #the 3600 comes from Normalised chp from p_el_max=3600. Look up chp_system_lib for more information
        self.P_el_max = config.p_el_max
        usually_P_el_max= config.P_el_max
        self.P_th_max = config.P_th_max*(self.P_el_max/usually_P_el_max)
        self.P_th_min = config.P_th_min
        self.P_el_min = config.P_el_min

        if self.P_el_max < self.P_el_min or self.P_th_max < self.P_th_min:
            self.P_el_max=self.P_el_min+100
            self.P_th_max=self.P_th_max+100

        self.mass_flow_max = config.mass_flow_max * (self.P_el_max/usually_P_el_max)
        if self.mass_flow_max < config.mass_flow_max:
            self.mass_flow_max=config.mass_flow_max
        self.eff_th_min:float = config.eff_th_min
        self.eff_th_max = config.eff_th_max
        self.eff_el_min:float = config.eff_el_min
        self.eff_el_max = config.eff_el_max
        self.temperature_max = config.temperature_max

        self.delta_T=config.delta_T

        #Inputs
        self.control_signal: ComponentInput = self.add_input(self.component_name, CHP.ControlSignal, lt.LoadTypes.ANY, lt.Units.PERCENT, False)
        #self.operating_mode_signal: ComponentInput = self.add_input(self.ComponentName, CHP.OperatingModelSignal, lt.LoadTypes.Gas, lt.Units.Percent, True)
        self.mass_inp_temp: ComponentInput = self.add_input(self.component_name, CHP.MassflowInputTemperature, lt.LoadTypes.WATER, lt.Units.CELSIUS, False)
        self.electricity_target: ComponentInput = self.add_input(self.component_name, CHP.ElectricityFromCHPTarget, lt.LoadTypes.ELECTRICITY, lt.Units.WATT, False)
        self.hydrogen_not_released: ComponentInput = self.add_input(self.component_name, CHP.HydrogenNotReleased, lt.LoadTypes.GAS, lt.Units.KG, False)


        #Outputs
        self.mass_out: ComponentOutput = self.add_output(self.component_name, CHP.MassflowOutput, lt.LoadTypes.WATER, lt.Units.KG_PER_SEC)
        self.mass_out_temp: ComponentOutput = self.add_output(self.component_name, CHP.MassflowOutputTemperature, lt.LoadTypes.WATER, lt.Units.CELSIUS)
        self.gas_demand_target: ComponentOutput = self.add_output(self.component_name, CHP.GasDemandTarget, lt.LoadTypes.GAS, lt.Units.KG_PER_SEC)
        self.el_power: ComponentOutput = self.add_output(self.component_name, CHP.ElectricityOutput, lt.LoadTypes.ELECTRICITY, lt.Units.WATT)
        self.number_of_cyclesC: ComponentOutput = self.add_output(self.component_name, CHP.NumberofCycles, lt.LoadTypes.ANY, lt.Units.ANY)
        self.th_power: ComponentOutput = self.add_output(self.component_name, CHP.ThermalOutputPower, lt.LoadTypes.HEATING, lt.Units.WATT)
        self.gas_demand_real_used: ComponentOutput = self.add_output(self.component_name, CHP.GasDemandReal, lt.LoadTypes.GAS, lt.Units.KG_PER_SEC)



    def i_prepare_simulation(self) -> None:
        """ Prepares the simulation. """
        pass

    @staticmethod
    def get_default_config() -> CHPConfig:
        config=CHPConfig(
                        name = "CHP",
                        min_operation_time = 60,
                        min_idle_time = 15,
                        gas_type = "Hydrogen",
                        operating_mode = "both",
                        p_el_max = 3600,
                        is_modulating = True,
                        P_el_min = 2_000,
                        P_th_min = 3_000,
                        eff_el_min = 0.2,
                        eff_th_min = 0.5,
                        mass_flow_max = 0.011,
                        P_el_max = 3_000,
                        P_th_max = 4_000,
                        eff_el_max = 0.4,
                        eff_th_max = 0.55,
                        temperature_max = 80,
                        delta_T = 10)
        return config
    def i_save_state(self) -> None:
        self.previous_state = copy.deepcopy(self.state)
        self.number_of_cycles_previous = self.number_of_cycles

    def i_restore_state(self) -> None:
        self.state = copy.deepcopy(self.previous_state)
        self.number_of_cycles = self.number_of_cycles_previous

    def i_doublecheck(self, timestep: int, stsv: SingleTimeStepValues) -> None:
         pass

    def simulate_chp(self,control_signal:float,stsv: SingleTimeStepValues,timestep:int)  -> Any:

        cw = 4182
        ## Calculation.Electric Energy deliverd
        ### CHP is on
        if self.state.activation != 0:
            self.number_of_cycles = self.state.cycle_number
            # Checks if the minimum running time has been reached

            # Minimium running time has been reached and the CHP wants to shut off -->so it shuts off
            if timestep >= self.state.start_timestep + self.min_operation_time and control_signal == 0.0:

                # all Outputs zero
                mass_out_temp:float = 0
                mass_out:float = 0
                el_power:float = 0
                th_power:float = 0
                eff_el_real:float = 0
                eff_th_real:float = 0

                self.state = CHPState(start_timestep=timestep,
                                      cycle_number=self.number_of_cycles,
                                      electricity_output=el_power)

                stsv.set_output_value(self.mass_out_temp, mass_out_temp)  # mass output temp
                stsv.set_output_value(self.mass_out, mass_out)  # mass output flow
                stsv.set_output_value(self.el_power, el_power)  # mass output flow
                stsv.set_output_value(self.th_power, th_power)  # mass output flow

                # zu ändern, da gas_demand=!gas_power
                stsv.set_output_value(self.number_of_cyclesC, self.number_of_cycles)
                return el_power, th_power, eff_el_real, eff_th_real



            # Minimium running time has not been reached and the CHP wants to shut off -->so it won't shut off
            elif timestep < self.state.start_timestep + self.min_operation_time and control_signal == 0:
                # CHP doesn't want to run but has to, therefore is going to run in minimum power

                maximum_power_th:float = self.P_th_min
                eff_th_real = self.eff_th_min
                eff_el_real = self.eff_el_min
                maximum_power_el:float = self.P_el_min

                th_power = maximum_power_th * eff_th_real
                el_power = maximum_power_el * eff_el_real

                mass_out_temp = self.delta_T + stsv.get_input_value(self.mass_inp_temp)
                mass_out = th_power / (cw * self.delta_T)

                if mass_out > self.mass_flow_max:
                    mass_out = self.mass_flow_max
                    mass_out_temp = stsv.get_input_value(self.mass_inp_temp) + th_power / (mass_out * cw)

            # CHP doens't want to shut off and its activated--> so its stays activated
            else:
                # Calculate Eff_th
                d_eff_th = (self.eff_th_max - self.eff_th_min)

                if control_signal * self.P_th_max < self.P_th_min:
                    maximum_power_th = self.P_th_min
                    eff_th_real = self.eff_th_min
                else:
                    maximum_power_th = control_signal * self.P_th_max
                    eff_th_real = self.eff_th_min + d_eff_th * control_signal

                # Calculate Eff_el
                d_eff_el = (self.eff_el_max - self.eff_el_min)

                if control_signal * self.P_el_max < self.P_el_min:
                    maximum_power_el = self.P_el_min
                    eff_el_real = self.eff_el_min
                else:
                    maximum_power_el = control_signal * self.P_el_max
                    eff_el_real = self.eff_el_min + d_eff_el * control_signal

                th_power = maximum_power_th * eff_th_real
                el_power = maximum_power_el * eff_el_real

                mass_out_temp = self.delta_T + stsv.get_input_value(self.mass_inp_temp)
                mass_out = th_power / (cw * self.delta_T)

                if mass_out > self.mass_flow_max:
                    mass_out = self.mass_flow_max
                    mass_out_temp = stsv.get_input_value(self.mass_inp_temp) + th_power / (mass_out * cw)

            th_power = (mass_out_temp - stsv.get_input_value(self.mass_inp_temp)) * cw * mass_out
            stsv.set_output_value(self.th_power, th_power)  # ThermalPowerOutput
            stsv.set_output_value(self.mass_out_temp, mass_out_temp)  # mass output temp
            stsv.set_output_value(self.mass_out, mass_out)  # mass output flow
            stsv.set_output_value(self.el_power, el_power)  # mass output flow
            # zu ändern, da gas_demand=!gas_power
            stsv.set_output_value(self.number_of_cyclesC, self.number_of_cycles)
            return el_power, th_power, eff_el_real, eff_th_real
            # run in power of control_signal

        ### CHP is Off
        # CHP wants to start and waited long enough since last start--> so it starts
        if control_signal != 0 and (timestep >= self.state.start_timestep + self.min_idle_time):
            self.number_of_cycles = self.number_of_cycles + 1
            number_of_cycles = self.number_of_cycles

            # Calculate Eff_th
            d_eff_th = (self.eff_th_max - self.eff_th_min)

            if control_signal * self.P_th_max < self.P_th_min:
                maximum_power_th = self.P_th_min
                eff_th_real = self.eff_th_min
            else:
                maximum_power_th = control_signal * self.P_th_max
                eff_th_real = self.eff_th_min + d_eff_th * control_signal

            # Calculate Eff_el
            d_eff_el = (self.eff_el_max - self.eff_el_min)

            if control_signal * self.P_el_max < self.P_el_min:
                maximum_power_el = self.P_el_min
                eff_el_real = self.eff_el_min
            else:
                maximum_power_el = control_signal * self.P_el_max
                eff_el_real = self.eff_el_min + d_eff_el * control_signal

            th_power = maximum_power_th * eff_th_real
            el_power = maximum_power_el * eff_el_real

            mass_out_temp = self.delta_T + stsv.get_input_value(self.mass_inp_temp)
            mass_out = th_power / (cw * self.delta_T)

            if mass_out > self.mass_flow_max:
                mass_out = self.mass_flow_max
                mass_out_temp = stsv.get_input_value(self.mass_inp_temp) + th_power / (mass_out * cw)

            self.state = CHPState(start_timestep=timestep,
                                  electricity_output=el_power,
                                  cycle_number=number_of_cycles)



        # CHP wants to starts but didn't wait long enough since last start -> so it won't start
        else:
            # all Outputs should be zero, because CHP can't start
            mass_out_temp = 0
            mass_out = 0
            el_power = 0
            th_power = 0
            eff_el_real = 0
            eff_th_real = 0

        stsv.set_output_value(self.th_power, th_power)  # ThermalPowerOutput
        stsv.set_output_value(self.mass_out_temp, mass_out_temp)  # mass output temp
        stsv.set_output_value(self.mass_out, mass_out)  # mass output flow
        stsv.set_output_value(self.el_power, el_power)  # mass output flow
        # zu ändern, da gas_demand=!gas_power
        stsv.set_output_value(self.number_of_cyclesC, self.number_of_cycles)

        return el_power, th_power, eff_el_real, eff_th_real

    def calculate_control_signal(self,stsv: SingleTimeStepValues)  -> float:
        if (stsv.get_input_value(self.electricity_target))<30:
            control_signal:float = 0
            return control_signal
        elif (stsv.get_input_value(self.electricity_target)) < self.P_el_min * self.eff_el_min:
            control_signal=0.4
            return control_signal
        elif (stsv.get_input_value(self.electricity_target))> self.P_el_max * self.eff_el_max:
            control_signal = 1
            return control_signal
        else:
            x1 = (-self.P_el_max - math.sqrt((self.P_el_max * self.eff_el_min) ** 2 + 4 * (
                        stsv.get_input_value(self.electricity_target) * self.P_el_max * (
                            self.eff_el_max - self.eff_el_min)))) / (
                             2 * self.P_el_max * (self.eff_el_max - self.eff_el_min))
            x2 = (-self.P_el_max + math.sqrt((self.P_el_max * self.eff_el_min) ** 2 + 4 * (
                        stsv.get_input_value(self.electricity_target) * self.P_el_max * (
                            self.eff_el_max - self.eff_el_min)))) / (
                             2 * self.P_el_max * (self.eff_el_max - self.eff_el_min))
            if 0<x1 and x1<1:
                if 0<x2 and x2<1:
                    if x1<x2:
                        return x2
                    else:
                        return x1
                else:
                    return x1
            else:
                return x2

    def i_simulate(self, timestep: int, stsv: SingleTimeStepValues,  force_convergence: bool) -> None:
        control_signal:float = -1
        if self.operating_mode=="heat":
            control_signal = stsv.get_input_value(self.control_signal)
        elif self.operating_mode=="electricity":
            control_signal = self.calculate_control_signal(stsv)
            if control_signal > 1 or control_signal<0:
                control_signal=1
        elif self.operating_mode=="both":
            control_signal=self.calculate_control_signal(stsv)
            if control_signal <= stsv.get_input_value(self.control_signal):
                    control_signal=stsv.get_input_value(self.control_signal)
            else:
                control_signal=control_signal
                if control_signal > 1 or control_signal<0:
                    control_signal=1

        if control_signal > 1:
            raise Exception("Expected a control signal between 0 and 1")
        if control_signal < 0:
            raise Exception("Expected a control signal between 0 and 1")


        el_power, th_power, eff_el_real, eff_th_real=self.simulate_chp(control_signal=control_signal, stsv=stsv, timestep=timestep)
        #Check if enough hydrogen is in the tank
        gas_demand_real_used = 0
        if el_power==0 and th_power==0:
            gas_demand_target=0
            gas_demand_real_used=0
        else:
            if self.gas_type == "Hydrogen":
                gas_demand_target= ((el_power/eff_el_real)+(th_power/eff_th_real ))/(PhysicsConfig.hydrogen_specific_fuel_value_per_kg)
                if stsv.get_input_value(self.hydrogen_not_released)==0:
                    #Gas Demand can completly be charged from storage
                    gas_demand_real_used=gas_demand_target

                elif stsv.get_input_value(self.hydrogen_not_released)<0:
                    log.warning("Fault, bc. GasDemandpossible>gasdemandtarget @ " + str(timestep))

                elif stsv.get_input_value(self.hydrogen_not_released)>0:
                    #not enough Gas for running CHP on power demanded/calculated before
                    #to simplify, turn of CHP complelty, also when minimum running time isn't reached, bec. no Hydrogen is there

                    stsv.set_output_value(self.th_power, 0)  # ThermalPowerOutput
                    stsv.set_output_value(self.mass_out_temp, 0)  # mass output temp
                    stsv.set_output_value(self.mass_out, 0)  # mass output flow
                    stsv.set_output_value(self.el_power, 0)  # mass output flow
                    self.number_of_cycles = self.state.cycle_number
                    el_power=0
                    self.state = CHPState(start_timestep=timestep,
                                          cycle_number=self.number_of_cycles,
                                          electricity_output=el_power)

                    stsv.set_output_value(self.number_of_cyclesC, self.number_of_cycles)
                    gas_demand_real_used=0
            elif self.gas_type == "Methan":
                gas_demand_target= ((el_power/eff_el_real)+(th_power/eff_th_real ))/PhysicsConfig.natural_gas_specific_fuel_value_per_kg
                gas_demand_real_used=gas_demand_target
            else:
                raise Exception("No Gas chosen which is integrated in System")


        stsv.set_output_value(self.gas_demand_target, gas_demand_target) #CHP runs with
        stsv.set_output_value(self.gas_demand_real_used, gas_demand_real_used)  # ThermalPowerOutput

    def write_to_report(self) -> List[str]:
        lines = []
        lines.append("CHP")
        # todo: add more useful stuff here
        lines.append("tbd")
        return lines


from typing import List
import pandas as pd
import os

from hisim import loadtypes as lt
from hisim import utils
from hisim import component as cp
from hisim.simulationparameters import SimulationParameters
from dataclasses import dataclass
from dataclasses_json import dataclass_json


@dataclass_json
@dataclass
class CSVLoaderConfig:
    component_name: str
    csv_filename: str
    column: int
    loadtype: lt.LoadTypes
    unit: lt.Units
    column_name: str
    sep: str
    decimal: str
    multiplier: float


class CSVLoader(cp.Component):
    """
    Class component loads CSV file containing some
    load profile relevant to the applied setup
    function

    Parameters
    --------------------------
    component_name: str
        Name of load profile from CSV file
    csv_filename: str
        Name of CSV filename containing the load profile data
    column: int
        Column number where the load profile data is stored
        inside of the CSV file
    loadtype: LoadTypes,
        Load type corresponded to the data loaded
    unit: lt.Units
        Units of data loaded
    column_name: str
        Name of column where the load profile data is stored
        inside of the CSV File
    simulation_parameters: cp.SimulationParameters
        Simulation parameters used by the setup function
    sep: str
        Separator used CSV file
    decimal: str
        Decimal indicator used in the CSV file
    multiplier: float
        Multiplication factor, in case an amplification of
        the data is required
    """

    Output1: str = "CSV Profile"

    def __init__(
        self, config: CSVLoaderConfig, my_simulation_parameters: SimulationParameters
    ):
        self.csvconfig = config
        super().__init__(
            name=self.csvconfig.component_name,
            my_simulation_parameters=my_simulation_parameters,
        )

        self.output1: cp.ComponentOutput = self.add_output(
            self.component_name,
            self.Output1,
            self.csvconfig.loadtype,
            self.csvconfig.unit,
        )
        self.output1.display_name = self.csvconfig.column_name
        self.multiplier = self.csvconfig.multiplier

        # ? self.column = column
        df = pd.read_csv(
            os.path.join(utils.HISIMPATH["inputs"], self.csvconfig.csv_filename),
            sep=self.csvconfig.sep,
            decimal=self.csvconfig.decimal,
        )
        if self.csvconfig.column >= len(df.columns):
            raise RuntimeError(
                f"Invalid column number for the csv file: {self.csvconfig.column}. Found {len(df.columns)} columns."
            )
        dfcolumn = df.iloc[:, [self.csvconfig.column]]
        self.column_name = self.csvconfig.column_name
        if len(dfcolumn) < self.my_simulation_parameters.timesteps:
            raise Exception(
                "Timesteps: "
                + str(self.my_simulation_parameters.timesteps)
                + " vs. Lines in CSV "
                + self.csvconfig.csv_filename
                + ": "
                + str(len(self.column_name))
            )

        self.column = dfcolumn.to_numpy(dtype=float)
        self.values: List[float] = []

    def i_restore_state(self) -> None:
        pass

    def i_simulate(
        self, timestep: int, stsv: cp.SingleTimeStepValues, force_convergence: bool
    ) -> None:
        stsv.set_output_value(
            self.output1, float(self.column[timestep]) * self.multiplier
        )

    def i_save_state(self) -> None:
        pass

    def i_doublecheck(self, timestep: int, stsv: cp.SingleTimeStepValues) -> None:
        pass



""" Contains all the main simulator components. """
# clean
import os
import datetime
from typing import List, Tuple, Optional, cast
import time

import pandas as pd

# Owned
from hisim.postprocessing.postprocessing_datatransfer import PostProcessingDataTransfer
from hisim.component_wrapper import ComponentWrapper
from hisim import sim_repository
from hisim.postprocessing import postprocessing_main as pp
import hisim.component as cp
from hisim import log
from hisim.simulationparameters import SimulationParameters
from hisim import utils


__authors__ = "Noah Pflugradt, Vitor Hugo Bellotto Zago, Maximillian Hillen"
__copyright__ = "Copyright 2020-2022, FZJ-IEK-3"
__license__ = "MIT"
__maintainer__ = "Noah Pflugradt"
__email__ = "n.pflugradt@fz-juelich.de"
__status__ = "production"


class Simulator:

    """ Core class of HiSim: Runs the main loop. """

    @utils.measure_execution_time
    def __init__(self, module_directory: str, setup_function: str, my_simulation_parameters: Optional[SimulationParameters]) -> None:
        """ Initializes the simulator class and creates the result directory. """
        if setup_function is None:
            raise Exception("No setup function was set")
        self.setup_function = setup_function
        self._simulation_parameters: SimulationParameters
        if my_simulation_parameters is not None:
            self._simulation_parameters = my_simulation_parameters
            log.LOGGING_LEVEL = self._simulation_parameters.logging_level
        self.wrapped_components: List[ComponentWrapper] = []
        self.all_outputs: List[cp.ComponentOutput] = []
        self.module_directory = module_directory
        self.simulation_repository = sim_repository.SimRepository()

    def set_simulation_parameters(self, my_simulation_parameters: SimulationParameters) -> None:
        """ Sets the simulation parameters and the logging level at the same time. """
        self._simulation_parameters = my_simulation_parameters
        if self._simulation_parameters is not None:
            log.LOGGING_LEVEL = self._simulation_parameters.logging_level

    def add_component(self, component: cp.Component, is_cachable: bool = False) -> None:
        """ Adds component to simulator and wraps it up the output in the register. """
        if self._simulation_parameters is None:
            raise Exception("Simulation Parameters were not initialized")
        # set the repository
        component.set_sim_repo(self.simulation_repository)

        # set the wrapper
        wrap = ComponentWrapper(component, is_cachable)
        wrap.register_component_outputs(self.all_outputs)
        self.wrapped_components.append(wrap)

    @utils.measure_execution_time
    def connect_all_components(self) -> None:
        """ Connects the inputs from every component to the corresponding outputs. """
        for wrapped_component in self.wrapped_components:
            wrapped_component.connect_inputs(self.all_outputs)

    @utils.measure_execution_time
    def prepare_calculation(self) -> None:
        """ Connects the inputs from every component to the corresponding outputs. """
        for wrapped_component in self.wrapped_components:
            wrapped_component.prepare_calculation()

    def process_one_timestep(self, timestep: int, previous_stsv: cp.SingleTimeStepValues) -> Tuple[cp.SingleTimeStepValues, int]:
        """ Executes one simulation timestep.

        Some components can be connected in a circle.
        To solve the circular dependency, all components have their states restored
        and simulated until their values converge.

        Firstly, their previously converged state is saved as the current timestep state.
        Following up, all components have their states restored and simulated respectively.
        Convergence is dependent on the i_restore and i_simulate of the components and how they
        are connected to each other.
        """

        # Save states of all components
        # Executes save state in the component
        for wrapped_component in self.wrapped_components:
            wrapped_component.save_state()

        continue_calculation = True

        # Verifies data existence
        if (len(self.all_outputs)) == 0:
            raise Exception("Not a single column was defined.")

        # Creates List with values
        stsv = previous_stsv.clone()
        # Creates a buffer List with values
        previous_values = previous_stsv.clone()
        iterative_tries = 0
        force_convergence = False

        # Starts loop
        while continue_calculation:
            # Loops through components
            for wrapped_component in self.wrapped_components:
                # Executes restore state for each component
                wrapped_component.restore_state()
                # Executes i_simulate for component
                wrapped_component.calculate_component(timestep, stsv, force_convergence)

            # Stops simulation for too small difference between
            # actual values and previous values
            if stsv.is_close_enough_to_previous(previous_values):
                continue_calculation = False

            if iterative_tries > 10:
                force_convergence = True
            if iterative_tries > 100:
                list_of_changed_values = stsv.get_differences_for_error_msg(previous_values, self.all_outputs)
                raise Exception("More than 100 tries in time step " + str(timestep) + "\n" + list_of_changed_values)
            # Copies actual values to previous variable
            previous_values.copy_values_from_other(stsv)
            iterative_tries += 1

        for wrapped_component in self.wrapped_components:
            wrapped_component.doublecheck(timestep, stsv)

        return (stsv, iterative_tries)

    def prepare_simulation_directory(self):
        """ Prepares the simulation directory. Determines the filename if nothing is set. """
        if self._simulation_parameters.result_directory is None or len(self._simulation_parameters.result_directory) == 0:
            result_dirname = f"{self.setup_function.lower()}_{datetime.datetime.now().strftime('%Y%m%d_%H%M%S')}"
            self._simulation_parameters.result_directory = os.path.join(self.module_directory, "results", result_dirname)

        if not os.path.isdir(self._simulation_parameters.result_directory):
            os.makedirs(self._simulation_parameters.result_directory, exist_ok=True)
        log.information("Using result directory: " + self._simulation_parameters.result_directory)
        log.LOGGING_LEVEL = self._simulation_parameters.logging_level

    # @profile
    # @utils.measure_execution_time
    def run_all_timesteps(self) -> None:
        """ Performs all the timesteps of the simulation and saves the results in the attribute results. """
        # Error Tests
        # Test if all parameters were initialized
        if self._simulation_parameters is None:
            raise Exception("Simulation Parameters were not initialized")

        # Tests if wrapper has any components at all
        if len(self.wrapped_components) == 0:
            raise Exception("Not a single component was defined. Quitting.")
        # call again because it might not have gotten executed depending on how it's called.
        self.prepare_simulation_directory()
        flagfile = os.path.join(self._simulation_parameters.result_directory, "finished.flag")
        if self._simulation_parameters.skip_finished_results and os.path.exists(flagfile):
            log.warning("Found " + flagfile + ". This calculation seems finished. Quitting.")
            return
        # Starts time counter
        start_counter = time.perf_counter()
        self.prepare_calculation()
        # Connects all components
        self.connect_all_components()
        log.information("finished connecting all components. A total of " + str(
            len(self.wrapped_components)) + " components were defined. They have a total of " + str(len(self.all_outputs)) + " outputs.")
        all_result_lines = []
        log.information("Starting simulation for " + str(self._simulation_parameters.timesteps) + " timesteps")
        lastmessage = datetime.datetime.now()
        last_step: int = 0
        starttime = datetime.datetime.now()
        total_iteration_tries_since_last_msg = 0

        # Creates empty list with values to get started
        number_of_outputs = len(self.all_outputs)
        stsv = cp.SingleTimeStepValues(number_of_outputs)

        for step in range(self._simulation_parameters.timesteps):
            if self._simulation_parameters.timesteps % 500 == 0:
                log.information("Starting step " + str(step))

            (resulting_stsv, iteration_tries) = self.process_one_timestep(step, stsv)
            stsv = cp.SingleTimeStepValues(number_of_outputs)
            # Accumulates iteration counter
            total_iteration_tries_since_last_msg += iteration_tries

            # Appends
            all_result_lines.append(resulting_stsv.values)
            del resulting_stsv
            # Calculates time execution
            elapsed = datetime.datetime.now() - lastmessage

            # For simulation longer than 5 seconds
            if elapsed.total_seconds() > 5:
                lastmessage = self.show_progress(starttime, step, total_iteration_tries_since_last_msg, last_step)
                last_step = step
                total_iteration_tries_since_last_msg = 0

        postprocessing_datatransfer = self.prepare_post_processing(all_result_lines, start_counter)
        log.information("Starting postprocessing")
        if postprocessing_datatransfer is None:
            raise Exception("postprocessing_datatransfer was none")

        my_post_processor = pp.PostProcessor()
        my_post_processor.run(ppdt=postprocessing_datatransfer)
        for wrapped_component in self.wrapped_components:
            wrapped_component.clear()
        del all_result_lines
        del postprocessing_datatransfer
        del my_post_processor
        self.simulation_repository.clear()
        log.information("Finished postprocessing")
        with open(flagfile, 'a', encoding="utf-8") as filestream:
            filestream.write("finished")

    @utils.measure_execution_time
    def prepare_post_processing(self, all_result_lines, start_counter):
        """ Prepares the post processing. """
        log.information("Preparing post processing")
        # Prepares the results from the simulation for the post processing.
        if len(all_result_lines) != self._simulation_parameters.timesteps:
            raise Exception("not all lines were generated")
        colum_names = []
        if self.setup_function is None:
            raise Exception("No setup function was set")
        entry: cp.ComponentOutput
        for _index, entry in enumerate(self.all_outputs):
            column_name = entry.get_pretty_name()
            colum_names.append(column_name)
            log.debug("Output column: " + column_name)
        results_data_frame = pd.DataFrame(data=all_result_lines, columns=colum_names)
        # todo: fix this constant
        df_index = pd.date_range("2021-01-01 00:00:00", periods=len(results_data_frame), freq="T")
        results_data_frame.index = df_index
        end_counter = time.perf_counter()
        execution_time = end_counter - start_counter
        log.information(f"Simulation took {execution_time:4.0f}s")
        results_merged = self.get_std_results(results_data_frame)

        ppdt = PostProcessingDataTransfer(
            results=results_data_frame,
            all_outputs=self.all_outputs,
            simulation_parameters=self._simulation_parameters,
            wrapped_components=self.wrapped_components,
            mode=1,
            setup_function=self.setup_function,
            execution_time=execution_time,
            results_monthly=results_merged,
        )
        log.information("Finished preparing post processing")
        return ppdt

    def show_progress(self, starttime: datetime.datetime, step: int, total_iteration_tries: int, last_step: int) -> datetime.datetime:
        """ Makes the pretty progress messages with time estimate. """
        # calculates elapsed time
        elapsed = datetime.datetime.now() - starttime
        elapsed_minutes, elapsed_seconds = divmod(elapsed.seconds, 60)
        elapsed_seconds_str: str = str(elapsed_seconds).zfill(2)
        # Calculates steps achieved per time duration
        steps_per_second = step / elapsed.total_seconds()
        elapsed_steps: int = step - last_step
        if elapsed_steps == 0:
            average_iteration_tries: float = 1
        else:
            average_iteration_tries = total_iteration_tries / elapsed_steps
        time_elapsed = datetime.timedelta(seconds=((self._simulation_parameters.timesteps - step) / steps_per_second))
        time_left_minutes, time_left_seconds = divmod(time_elapsed.seconds, 60)
        time_left_seconds = str(time_left_seconds).zfill(2)  # type: ignore
        simulation_status = f"Simulating... {(step / self._simulation_parameters.timesteps) * 100:.1f}% "
        simulation_status += f"| Elapsed Time: {elapsed_minutes}:{elapsed_seconds_str} min "
        simulation_status += f"| Speed: {steps_per_second:.0f} step/s "
        simulation_status += f"| Time Left: {time_left_minutes}:{time_left_seconds} min"
        simulation_status += f"| Avg. iterations {average_iteration_tries:.1f}"
        log.information(simulation_status)
        return datetime.datetime.now()

    def get_std_results(self, results_data_frame: pd.DataFrame) -> pd.DataFrame:
        """ Converts results into a pretty dataframe for post processing. """
        pd_timeline = pd.date_range(start=self._simulation_parameters.start_date,
                                    end=self._simulation_parameters.end_date,
                                    freq=f'{self._simulation_parameters.seconds_per_timestep}S')[:-1]
        n_columns = results_data_frame.shape[1]
        results_data_frame.index = pd_timeline
        results_merged = pd.DataFrame()
        for i_column in range(n_columns):
            temp_df = pd.DataFrame(results_data_frame.values[:, i_column], index=pd_timeline,
                                   columns=[results_data_frame.columns[i_column]])
            column_name1 = results_data_frame.columns[i_column]  # noqa
            column_name: str = cast(str, column_name1)
            if 'Temperature' in column_name or 'Percent' in column_name:
                temp_df = temp_df.resample('M').interpolate(method='linear')
            else:
                temp_df = temp_df.resample('M').sum()
            results_merged[temp_df.columns[0]] = temp_df.values[:, 0]
            results_merged.index = temp_df.index
        return results_merged



""" Enum classes to help against string constants.

Guidelines for enum classes:
    1. Write members names extensively, with no abbreviation, i.e., 'Watt' instead of 'W'.
    2. Attributes should follow the International System of Units (SI)
    [https://en.wikipedia.org/wiki/International_System_of_Units], i.e., for power the attribute is 'W'.
    3. Do not use multipliers such as 'Kilowatt'.
        3.1 Exceptions to this rule are: 'Kilometer', 'Kilogram'.
        3.2 In case of an exception, the simple form should be avoided altogether, e.g., given the 'Kilometer'
        is an Unit, then 'Meter' should not be used.

"""
# clean
import enum


@enum.unique
class Locations(str, enum.Enum):

    """ To parse locations for climate data. """

    AACHEN = "Aachen"


@enum.unique
class ChargingLocations(str, enum.Enum):

    """ To parse locations for EV Charging. """

    AT_HOME = "AtHome"
    AT_WORK = "AtWork"


@enum.unique
class OccupancyProfiles(str, enum.Enum):

    """ To parse LPG predefined household strings. """

    CH01 = "CH01"


class BuildingCodes(str, enum.Enum):

    """ To parse predefined house types from tabula. """

    DE_N_SFH_05_GEN_REEX_001_002 = "DE.N.SFH.05.Gen.ReEx.001.002"


@enum.unique
class DisplayNames(str, enum.Enum):

    """ For the sankey plotting. """

    ELECTRICITY_OUTPUT = "ElectricityOutput"
    ELECTRICITY_INPUT = "ElectricityInput"


@enum.unique
class Termination(str, enum.Enum):

    """ For the simulation status of modular household. """

    SUCCESSFUL = "Sucessful"
    INVESTMENT_EXCEEDED = "InvestmentExceeded"


@enum.unique
class LoadTypes(str, enum.Enum):

    """ Load type named constants so that they are the same everywhere and no typos happen. """

    ANY = "Any"

    ELECTRICITY = "Electricity"
    IRRADIANCE = "Irradiance"
    SPEED = "Speed"
    HEATING = "Heating"
    COOLING = "Cooling"

    VOLUME = "Volume"
    TEMPERATURE = "Temperature"
    TIME = "Time"

    # Substance
    GAS = "Gas"
    HYDROGEN = "Hydrogen"
    OXYGEN = "Oxygen"
    WATER = "Water"
    WARM_WATER = "WarmWater"
    DIESEL = "Diesel"
    OIL = "Oil"
    DISTRICTHEATING = "DistrictHeating"

    PRICE = "Price"

    # Controllers:
    ON_OFF = "OnOff"  # encoding: 0 means off and 1 means on
    ACTIVATION = 'Activation'


@enum.unique
class Units(str, enum.Enum):

    """ Unit Constants. """

    # Unphysical
    ANY = "-"
    PERCENT = "%"

    # Power
    WATT = "W"
    KILOWATT = "kW"
    KWH_PER_TIMESTEP = "kWh per timestep"

    # Power per area
    WATT_PER_SQUARE_METER = "W per square meter"
    WATT_HOUR_PER_SQUARE_METER = "Wh per square meter"

    # Speed
    METER_PER_SECOND = "m/s"

    # Energy
    WATT_HOUR = "Wh"
    KWH = "kWh"

    # Volume
    LITER = "L"

    # Volume per time
    LITER_PER_TIMESTEP = "Liter per timestep"

    # Mass
    KG = "kg"

    # Mass flow
    KG_PER_SEC = "kg/s"

    # Degrees
    CELSIUS = "°C"
    KELVIN = 'K'

    # Degrees
    DEGREES = "Degrees"

    # Time
    SECONDS = "s"
    TIMESTEPS = 'timesteps'
    YEARS = 'years'

    # Cost
    EUR_PER_KWH = "Euros per kWh"
    EURO = "Euro"

    # Binary for controllers
    BINARY = 'binary'


@enum.unique
class ComponentType(str, enum.Enum):

    """ Component types for use in dynamic controllers. """

    PV = "PV"
    SMART_DEVICE = "SmartDevice"
    SURPLUS_CONTROLLER = "SurplusController"
    PREDICTIVE_CONTROLLER = "PredictiveControllerforSmartDevices"
    HEAT_PUMP = "HeatPump"
    GAS_HEATER = "GasHeater"
    BATTERY = "Battery"
    CAR_BATTERY = "CarBattery"
    FUEL_CELL = "FuelCell"
    ELECTROLYZER = "Electrolyzer"
    CHP = "CHP"
    H2_STORAGE = "H2Storage"
    ELECTRIC_VEHICLE = "ElectricVehicle"
    ELECTRIC_BOILER = "ElectricBoiler"
    BOILER = "Boiler"
    BUFFER = "Buffer"
    HEATERS = [HEAT_PUMP, GAS_HEATER]


@enum.unique
class InandOutputType(str, enum.Enum):

    """ For dynamic controllers. """

    MASS_FLOW = "Massflow"
    CONTROL_SIGNAL = "ControlSignal"
    ELECTRICITY_TARGET = "ElectricityTarget"
    ELECTRICITY_REAL = "ElectricityReal"

    # L3
    LAST_ACTIVATION = "LastActivation"
    LATEST_ACTIVATION = "LatestActivation"
    EARLIEST_ACTIVATION = "EarliestActivation"
    RECOMMENDED_ACTIVATION = "RecommendedActivation"

    # Energy Management System / Postprocessing Options
    ELECTRICITY_PRODUCTION = "ElectricityProduction"
    ELECTRICITY_INJECTION = "ElectricityInjection"
    ELECTRICITY_CONSUMPTION = "ElectricityConsumption"
    ELECTRICITY_CONSUMPTION_EMS_CONTROLLED = "Consumption with EMS control"
    ELECTRICITY_CONSUMPTION_UNCONTROLLED = "Consumption without any EMS control"
    STORAGE_CONTENT = "StorageContent"
    CHARGE_DISCHARGE = "ChargeDischarge"

    # Heating
    HEAT_TO_BUILDING = "HeatToBuilding"
    HEAT_TO_BUFFER = "HeatToBuffer"
    THERMAL_PRODUCTION = "ThermalProduction"
    FUEL_CONSUMPTION = "FuelConsumption"



# -*- coding: utf-8 -*-
""" Controller of EV battery. """

from typing import List, Any
from dataclasses import dataclass
from dataclasses_json import dataclass_json

from utspclient.helpers.lpgpythonbindings import JsonReference
from utspclient.helpers.lpgdata import ChargingStationSets

from hisim.simulationparameters import SimulationParameters
from hisim import component as cp
from hisim import loadtypes as lt
from hisim import log
from hisim.components import generic_car
from hisim.components import advanced_ev_battery_bslib

__authors__ = "Johanna Ganglbauer"
__copyright__ = "Copyright 2021, the House Infrastructure Project"
__credits__ = ["Noah Pflugradt"]
__license__ = ""
__version__ = ""
__maintainer__ = "Johanna Ganglbauer"
__email__ = "johanna.ganglbauer@4wardenergy.at"
__status__ = "development"


@dataclass_json
@dataclass
class ChargingStationConfig:

    """ Definition of the configuration of Charging Station. """

    name: str
    source_weight: int
    charging_station_set: JsonReference
    battery_set: float

    def __init__(self, name: str, source_weight: int, charging_station_set: JsonReference,
                 battery_set: float) -> None:
        """ Initialization of the configuration of Charging Station. """

        self.name = name
        self.source_weight = source_weight
        self.charging_station_set = charging_station_set
        self.battery_set = battery_set

    @staticmethod
    def get_default_config(charging_station_set: JsonReference = ChargingStationSets.Charging_At_Home_with_03_7_kW) -> Any:
        """ Returns default configuration of charging station and desired SOC Level. """
        config = ChargingStationConfig(
            name='L1EVChargeControl', source_weight=1, charging_station_set=charging_station_set,
            battery_set=0.8)
        return config


class L1ControllerState:

    """Data class, which saves the state of the controller. """

    def __init__(self, power: float) -> None:
        """ Initializes power for battery charge/discharge in state. """
        self.power = power

    def clone(self) -> Any:
        """ Copy state efficiently. """
        return L1ControllerState(power=self.power)


class L1Controller(cp.Component):

    """ Simulates EV charging with constant SOC threshold and optional surplus control. """

    # Inputs
    ElectricityOutput = "ElectricityOutput"
    CarLocation = "CarLocation"
    StateOfCharge = "StateOfCharge"
    ElectricityTarget = "ElectricityTarget"

    # Outputs
    ToOrFromBattery = "ToOrFromBattery"

    def __init__(self, my_simulation_parameters: SimulationParameters, config: ChargingStationConfig) -> None:
        """ Initializes Car. """
        super().__init__(name=config.name + '_w' + str(config.source_weight),
                         my_simulation_parameters=my_simulation_parameters)
        self.state = L1ControllerState(power=0)
        self.previous_state = L1ControllerState(power=0)
        self.build(config=config, my_simulation_parameters=my_simulation_parameters)

        # add inputs
        self.car_consumption: cp.ComponentInput = self.add_input(
            self.component_name, self.ElectricityOutput, lt.LoadTypes.ELECTRICITY, lt.Units.WATT,
            mandatory=True)
        self.car_location: cp.ComponentInput = self.add_input(
            self.component_name, self.CarLocation, lt.LoadTypes.ANY, lt.Units.ANY,
            mandatory=True)
        self.state_of_charge: cp.ComponentInput = self.add_input(
            self.component_name, self.StateOfCharge, lt.LoadTypes.ANY, lt.Units.ANY,
            mandatory=True)

        if self.clever:
            self.electricity_target: cp.ComponentInput = self.add_input(
                self.component_name, self.ElectricityTarget, lt.LoadTypes.ELECTRICITY, lt.Units.WATT,
                mandatory=True)

        # add outputs
        self.p_set: cp.ComponentOutput = self.add_output(
            object_name=self.component_name, field_name=self.ToOrFromBattery, load_type=lt.LoadTypes.ELECTRICITY,
            unit=lt.Units.WATT)

        self.add_default_connections(generic_car.Car, self.get_car_default_connections())
        self.add_default_connections(advanced_ev_battery_bslib.CarBattery, self.get_battery_default_connections())

    def get_car_default_connections(self) -> List[cp.ComponentConnection]:
        """ Default connections of car in ev charge controller. """
        log.information("setting car default connections in ev charge controler")
        connections: List[cp.ComponentConnection] = []
        car_classname = generic_car.Car.get_classname()
        connections.append(cp.ComponentConnection(L1Controller.ElectricityOutput, car_classname, generic_car.Car.ElectricityOutput))
        connections.append(cp.ComponentConnection(L1Controller.CarLocation, car_classname, generic_car.Car.CarLocation))
        return connections

    def get_battery_default_connections(self) -> List[cp.ComponentConnection]:
        """ Default connections of battery in ev charge controller. """
        log.information("setting battery default connections in ev charge controler")
        connections: List[cp.ComponentConnection] = []
        battery_classname = advanced_ev_battery_bslib.CarBattery.get_classname()
        connections.append(cp.ComponentConnection(L1Controller.StateOfCharge, battery_classname, advanced_ev_battery_bslib.CarBattery.StateOfCharge))
        return connections

    def i_save_state(self) -> None:
        """ Saves actual state. """
        self.previous_state = self.state.clone()

    def i_restore_state(self) -> None:
        """ Restores previous state. """
        self.state = self.previous_state.clone()

    def i_doublecheck(self, timestep: int, stsv: cp.SingleTimeStepValues) -> None:
        """ Checks statements. """
        pass

    def i_prepare_simulation(self) -> None:
        """ Prepares the simulation. """
        pass

    def i_simulate(self, timestep: int, stsv: cp.SingleTimeStepValues,  force_convergence: bool) -> None:
        """ Returns battery charge and discharge (energy consumption of car) of battery at each timestep. """
        if force_convergence:
            pass
        else:
            car_location = stsv.get_input_value(self.car_location)
            car_consumption = stsv.get_input_value(self.car_consumption)
            soc = stsv.get_input_value(self.state_of_charge)

            if car_consumption > 0:
                self.state.power = -car_consumption
            elif soc < self.battery_set and car_location == self.charging_location:
                self.state.power = self.power
            else:
                self.state.power = 0
                if self.clever:
                    electricity_target = stsv.get_input_value(self.electricity_target)
                    if electricity_target > 0:
                        self.state.power = electricity_target

            stsv.set_output_value(self.p_set, self.state.power)

    def build(self, config: ChargingStationConfig, my_simulation_parameters: SimulationParameters) -> None:
        """ Translates and assigns config parameters to controller class and completes initialization. """
        self.name = config.name
        self.source_weight = config.source_weight
        # get charging station location and charging station power out of ChargingStationSet
        if config.charging_station_set.Name is not None:
            charging_station_string = config.charging_station_set.Name.partition('At ')[2]
            location = charging_station_string.partition(' with')[0]
            if location == 'Home':
                self.charging_location = 1
            elif location == 'Work':
                self.charging_location = 2
        else:
            log.error('Charging location not known, check the input on the charging station set. It was set to "charging at home per default.')
        power = float(charging_station_string.partition('with ')[2].partition(' kW')[0]) * 1e3
        self.power = power
        self.battery_set = config.battery_set
        self.clever = my_simulation_parameters.system_config.clever

    def write_to_report(self) -> List[str]:
        """ Writes EV charge controller values to report. """
        lines = []
        lines.append(self.name + '_w' + str(self.source_weight) + 'charging controller: ')
        lines.append(f"Power [kW]: {self.power * 1e-3:2.1f}")
        if self.charging_location == 1:
            lines.append("At Home")
        elif self.charging_location == 2:
            lines.append("At Work")
        return lines



# -*- coding: utf-8 -*-
from typing import Optional
from dataclasses import dataclass
from dataclasses_json import dataclass_json

from utspclient.helpers.lpgdata import TransportationDeviceSets, TravelRouteSets, Households
from utspclient.helpers.lpgpythonbindings import JsonReference
from hisim.loadtypes import Locations, BuildingCodes

@dataclass_json
@dataclass
class ArcheTypeConfig:

    """ Defines the system config for the modular household. """

    location: Locations
    occupancy_profile: JsonReference
    building_code: BuildingCodes
    mobility_set: JsonReference
    mobility_distance: Optional[JsonReference]

    def __init__(self, location: Locations = Locations.AACHEN,
            occupancy_profile: JsonReference = Households.CHR01_Couple_both_at_Work,
            building_code: BuildingCodes = BuildingCodes.DE_N_SFH_05_GEN_REEX_001_002,
            mobility_set: JsonReference = TransportationDeviceSets.Bus_and_two_30_km_h_Cars,
            mobility_distance: Optional[JsonReference] = TravelRouteSets.Travel_Route_Set_for_10km_Commuting_Distance):  # noqa
        self.location = location
        self.occupancy_profile = occupancy_profile
        self.building_code = building_code
        self.mobility_set = mobility_set
        self.mobility_distance = mobility_distance

# def create_archetype_config_file() -> None:
#     """Component Cost file is created."""

#     config_file=ArcheTypeConfig()
#     config_file_written = config_file.to_json()

#     with open('arche_type_config.json', 'w', encoding="utf-8") as outfile:
#         outfile.write(config_file_written)



""" For calling a model with profiler and creating a log file. """
# clean
import cProfile
import pstats
import hisim.hisim_main as hsm


def maincall() -> None:
    """ For calling the Hisim main. """
    # change call here as needed
    hsm.main("..\\examples\\modular_example.py", "modular_household_explicit")


if __name__ == "__main__":
    """ Called from the command line.
    This function calls HiSim main and performs a profiling with cprofile.
    The results are dumped to various text files in the result directory
    and the .prof file can be visualized with for example snakeviz.
    """

    profiler = cProfile.Profile()
    profiler.enable()
    maincall()
    profiler.disable()

    with open("..\\examples\\results\\profilingStatsAsTextSortedCumulative.txt", "w", encoding="utf-8") as f:
        stats = pstats.Stats(profiler, stream=f).sort_stats('cumulative')
        stats.print_stats()
    with open("..\\examples\\results\\profilingStatsAsTextSortedcalls.txt", "w", encoding="utf-8") as f:
        stats = pstats.Stats(profiler, stream=f).sort_stats('ncalls')
        stats.print_stats()
    with open("..\\examples\\results\\profilingStatsAsTextSortedTotalTime.txt", "w", encoding="utf-8") as f:
        stats = pstats.Stats(profiler, stream=f).sort_stats('tottime')
        stats.print_stats()
    stats.dump_stats('..\\examples\\results\\profile-export-data.prof')
